# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import pytest
import scipp as sc
from scipp.testing import assert_identical

from scippneutron.tof import fakes


@pytest.fixture
def ess_10s_14Hz() -> fakes.FakeSource:
    return fakes.FakeSource(
        frequency=sc.scalar(14.0, unit='Hz'), run_length=sc.scalar(10.0, unit='s')
    )


def test_fake_source(ess_10s_14Hz) -> None:
    source = ess_10s_14Hz
    assert source.t0.unit == sc.Unit('ns')
    assert source.t0.sizes == {'pulse': 140}
    assert_identical(source.t0[0], sc.datetime('2019-12-25T06:00:00.0', unit='ns'))
    assert_identical(
        source.t0[-1],
        sc.datetime('2019-12-25T06:00:00.0', unit='ns')
        + sc.scalar(139 / 140 * 10 * 1e9, unit='ns', dtype='int64'),
    )


def test_fake_monitor(ess_10s_14Hz) -> None:
    pulse = fakes.FakePulse(
        time_min=sc.scalar(0.0, unit='ms'),
        time_max=sc.scalar(3.0, unit='ms'),
        wavelength_min=sc.scalar(0.1, unit='angstrom'),
        wavelength_max=sc.scalar(10.0, unit='angstrom'),
    )
    beamline = fakes.FakeBeamline(
        source=ess_10s_14Hz,
        pulse=pulse,
        choppers=fakes.wfm_choppers,
        monitors={'monitor': sc.scalar(26.0, unit='m')},
        detectors={},
    )
    mon, _ = beamline.get_monitor('monitor')
    assert mon.sizes == {'pulse': 140}



from typing import Any, NewType, Optional

import sphinx

_typehint_aliases = {
    'scipp._scipp.core.DataArray': 'scipp.DataArray',
    'scipp._scipp.core.Dataset': 'scipp.Dataset',
    'scipp._scipp.core.DType': 'scipp.DType',
    'scipp._scipp.core.Unit': 'scipp.Unit',
    'scipp._scipp.core.Variable': 'scipp.Variable',
    'scipp.core.data_group.DataGroup': 'scipp.DataGroup',
}


def typehints_formatter_for(package: str) -> callable:
    def typehints_formatter(annotation, config: sphinx.config.Config) -> Optional[str]:
        """Format typehints with improved NewType handling."""
        _ = config
        if _is_new_type(annotation):
            return _format_new_type(annotation, package)
        if _is_type_alias_type(annotation):
            return _format_type_alias_type(annotation, package)
        return None

    return typehints_formatter


def _is_new_type(annotation: Any) -> bool:
    # TODO Switch to isinstance(key, NewType) once our minimum is Python 3.10
    # Note that we cannot pass mypy in Python<3.10 since NewType is not a type.
    return hasattr(annotation, '__supertype__')


def _format_new_type(annotation: NewType, package: str) -> str:
    return (
        f'{_internal_link(annotation, "class", package)}'
        f' ({_link(annotation.__supertype__, "class")})'
    )


def _is_type_alias_type(annotation) -> bool:
    try:
        from typing import TypeAliasType

        return isinstance(annotation, TypeAliasType)
    except ImportError:
        return False  # pre python 3.12


def _format_type_alias_type(annotation: Any, package: str) -> str:
    alias = _internal_link(annotation, "class", package, annotation.__type_params__)
    value = _link(annotation.__value__, "class", _get_type_args(annotation.__value__))
    return f'{alias} ({value})'


def _get_type_args(ty: type) -> tuple[type, ...]:
    if (args := getattr(ty, '__args__', None)) is not None:
        return args  # e.g. list[int]
    return ty.__type_params__


def _internal_link(
    annotation: Any,
    kind: str,
    package: str,
    type_params: Optional[tuple[type, ...]] = None,
) -> str:
    target = f'{annotation.__module__}.{annotation.__name__}'
    label = f'{annotation.__module__.removeprefix(package+".")}.{annotation.__name__}'
    if type_params:
        label += f'[{", ".join(ty.__name__ for ty in type_params)}]'
    return f':{kind}:`{label} <{target}>`'


def _link(ty: type, kind: str, type_params: Optional[tuple[type, ...]] = None) -> str:
    if ty.__module__ == 'builtins':
        target = ty.__name__
    else:
        target = f'{ty.__module__}.{ty.__name__}'
    label = _typehint_aliases.get(target, target)
    if type_params:
        label += f'[{", ".join(ty.__name__ for ty in type_params)}]'
    return f':{kind}:`{label} <{target}>`'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""Graphs for computing coordinates in time-of-flight neutron scattering.

See :mod:`scippneutron.conversion.tof` for definitions
of the quantities used here.
And the `user guide <../../user-guide/coordinate-transformations.rst>`_
for examples.

Functions in this module come in two categories and return graphs that

- can be used to compute a specific coordinate as identified by the function name
  (e.g. ``elastic_energy``) from a given coordinate as given by the ``start`` argument.
  These graphs may contain more than one node if necessary.
- can be used to compute multiple coordinates, (``elastic`` and ``kinematic``).
  Their ``start`` argument works as in the other functions.
"""

from typing import Callable, Dict

from .. import tof as _kernels

Graph = Dict[str, Callable]

_GRAPH_DYNAMICS_BY_ORIGIN = {
    'energy': {
        'dspacing': _kernels.dspacing_from_energy,
        'wavelength': _kernels.wavelength_from_energy,
    },
    'tof': {
        'dspacing': _kernels.dspacing_from_tof,
        'energy': _kernels.energy_from_tof,
        'hkl_vec': _kernels.hkl_vec_from_Q_vec,
        ('h', 'k', 'l'): _kernels.hkl_elements_from_hkl_vec,
        'ub_matrix': _kernels.ub_matrix_from_u_and_b,
        'Q': _kernels.Q_from_wavelength,
        'Q_vec': _kernels.Q_vec_from_Q_elements,
        ('Qx', 'Qy', 'Qz'): _kernels.Q_elements_from_wavelength,
        'wavelength': _kernels.wavelength_from_tof,
        'time_at_sample': _kernels.time_at_sample_from_tof,
    },
    'Q': {
        'wavelength': _kernels.wavelength_from_Q,
    },
    'wavelength': {
        'dspacing': _kernels.dspacing_from_wavelength,
        'energy': _kernels.energy_from_wavelength,
        'hkl_vec': _kernels.hkl_vec_from_Q_vec,
        ('h', 'k', 'l'): _kernels.hkl_elements_from_hkl_vec,
        'ub_matrix': _kernels.ub_matrix_from_u_and_b,
        'Q': _kernels.Q_from_wavelength,
        'Q_vec': _kernels.Q_vec_from_Q_elements,
        ('Qx', 'Qy', 'Qz'): _kernels.Q_elements_from_wavelength,
    },
}


def _strip_elastic(start: str, keep: list) -> Graph:
    full_graph = elastic(start)
    return {key: full_graph[key] for key in keep if key != start}


def elastic(start: str) -> Graph:
    """Graph for elastic scattering transformations.

    Parameters
    ----------
    start:
        Input coordinate. One of 'dspacing', 'energy', 'tof', 'Q',
        or 'wavelength'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return dict(_GRAPH_DYNAMICS_BY_ORIGIN[start])


def kinematic(start: str) -> Graph:
    """Graph with pure kinematics.

    The returned graph can be used to compute scattering-independent quantities.

    Parameters
    ----------
    start:
        Input coordinate. Currently, only 'tof' is supported.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(start, keep=['wavelength', 'energy'])


def elastic_dspacing(start: str) -> Graph:
    """
    Graph for elastic scattering transformation to dspacing.

    Parameters
    ----------
    start:
     Input coordinate. One of 'energy', 'tof', or 'wavelength'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(start, keep=['dspacing'])


def elastic_energy(start: str) -> Graph:
    """
    Graph for elastic scattering transformation to energy.

    Parameters
    ----------
    start:
        Input coordinate. One of 'tof' or 'wavelength'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(start, keep=['energy'])


def elastic_Q(start: str) -> Graph:
    """
    Graph for elastic scattering transformation to Q.

    Parameters
    ----------
    start:
        Input coordinate. One of 'tof' or 'wavelength'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(start, keep=['Q', 'wavelength'])


def elastic_Q_vec(start: str) -> Graph:
    """
    Graph for elastic scattering transformation to Q vector.

    Parameters
    ----------
    start:
        Input coordinate. One of 'tof' or 'wavelength'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(start, keep=[('Qx', 'Qy', 'Qz'), 'Q_vec', 'wavelength'])


def elastic_hkl(start: str) -> Graph:
    """
    Graph for elastic scattering transformation to Q vector.

    Parameters
    ----------
    start:
        Input coordinate. One of 'tof' or 'wavelength'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(
        start,
        keep=[
            ('Qx', 'Qy', 'Qz'),
            'Q_vec',
            ('h', 'k', 'l'),
            'hkl_vec',
            'ub_matrix',
            'wavelength',
        ],
    )


def elastic_wavelength(start: str) -> Graph:
    """
    Graph for elastic scattering transformation to wavelength.

    Parameters
    ----------
    start:
        Input coordinate. One of 'energy', 'tof', or 'Q'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return _strip_elastic(start, keep=['wavelength'])


def direct_inelastic(start: str) -> Graph:
    """
    Graph for direct-inelastic scattering transformations.

    Parameters
    ----------
    start:
        Input coordinate. Currently, only 'tof' is supported.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return {'tof': {'energy_transfer': _kernels.energy_transfer_direct_from_tof}}[start]


def indirect_inelastic(start: str) -> Graph:
    """
    Graph for indirect-inelastic scattering transformations.

    Parameters
    ----------
    start:
        Input coordinate. Currently, only 'tof' is supported.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return {'tof': {'energy_transfer': _kernels.energy_transfer_indirect_from_tof}}[
        start
    ]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from enum import Enum


class ConsumerType(Enum):
    """
    Picklable way of instructing the mp.Process to
    construct real consumers or fakes for tests
    """

    REAL = 1
    FAKE = 2



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import pytest

from scippneutron.conversion.graph import beamline


def test_incident_beam_has_correct_keys():
    assert set(beamline.incident_beam().keys()) == {'incident_beam'}


def test_scattered_beam_has_correct_keys():
    assert set(beamline.scattered_beam().keys()) == {'scattered_beam'}


def test_two_theta_has_correct_keys():
    assert set(beamline.two_theta().keys()) == {
        'scattered_beam',
        'incident_beam',
        'two_theta',
    }


def test_L1_has_correct_keys():
    assert set(beamline.L1().keys()) == {'L1', 'incident_beam'}


def test_L2_has_correct_keys():
    assert set(beamline.L2().keys()) == {'L2', 'scattered_beam'}


def test_Ltotal_has_correct_keys():
    assert set(beamline.Ltotal(scatter=False).keys()) == {'Ltotal'}
    assert set(beamline.Ltotal(scatter=True).keys()) == {
        'scattered_beam',
        'incident_beam',
        'L1',
        'L2',
        'Ltotal',
    }


def test_beamline_has_correct_keys():
    assert set(beamline.beamline(scatter=False).keys()) == {'Ltotal'}
    assert set(beamline.beamline(scatter=True).keys()) == {
        'scattered_beam',
        'incident_beam',
        'L1',
        'L2',
        'Ltotal',
        'two_theta',
    }


@pytest.mark.parametrize(
    'fn',
    (
        beamline.incident_beam,
        beamline.scattered_beam,
        beamline.L1,
        beamline.L2,
        beamline.two_theta,
    ),
)
def test_beamline_returns_new_graph_without_scatter_arg(fn):
    g = fn()
    g['a_new_node'] = lambda position: 2 * position
    assert 'a_new_node' not in fn()


@pytest.mark.parametrize('fn', (beamline.beamline, beamline.Ltotal))
@pytest.mark.parametrize('scatter', (True, False))
def test_beamline_returns_new_graph_with_scatter_arg(fn, scatter):
    g = fn(scatter=scatter)
    g['a_new_node'] = lambda position: 2 * position
    assert 'a_new_node' not in fn(scatter=scatter)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import asyncio
import multiprocessing as mp
import time
from enum import Enum
from queue import Empty as QueueEmpty
from typing import Generator, List, Optional
from warnings import warn

import numpy as np
import scipp as sc

from ..io.nexus.load_nexus import load_nexus_json_str
from ._consumer_type import ConsumerType
from ._data_stream_widget import DataStreamWidget
from ._serialisation import convert_from_pickleable_dict
from ._stop_time import StopTimeUpdate


class StartTime(Enum):
    NOW = "now"
    START_OF_RUN = "start_of_run"


class StopTime(Enum):
    NEVER = "never"
    END_OF_RUN = "end_of_run"


_missing_dependency_message = (
    "Confluent Kafka Python library and/or serialisation library"
    "not found, please install confluent-kafka and "
    "ess-streaming-data-types as detailed in the "
    "installation instructions (https://scipp.github.io/"
    "scippneutron/getting-started/installation.html)"
)


async def data_stream(
    kafka_broker: str,
    topics: Optional[List[str]] = None,
    event_buffer_size: int = 1_048_576,
    slow_metadata_buffer_size: int = 1000,
    fast_metadata_buffer_size: int = 100_000,
    chopper_buffer_size: int = 10_000,
    interval: sc.Variable = 2.0 * sc.units.s,
    run_info_topic: Optional[str] = None,
    start_time: StartTime = StartTime.NOW,
    stop_time: StopTime = StopTime.NEVER,
) -> Generator[sc.DataArray, None, None]:
    """
    Periodically yields accumulated data from stream.
    If the buffer fills up more frequently than the set interval
    then data is yielded more frequently.
    1048576 event buffer is around 24 MB (with pulse_time, id, weights, etc)
    :param kafka_broker: Address of the Kafka broker to stream data from
    :param topics: Kafka topics to consume data from (not required if
      run_info_topic is used)
    :param event_buffer_size: Size of buffer to accumulate event data in
    :param slow_metadata_buffer_size: Size of buffer to accumulate slow
      sample env metadata in
    :param fast_metadata_buffer_size: Size of buffer to accumulate fast
      sample env metadata in
    :param chopper_buffer_size: Size of buffer to accumulate chopper
      timestamps in
    :param interval: interval between yielding any new data
      collected from stream
    :param run_info_topic: If provided, the first data batch returned by
      data_stream will be from the last available run start message in
      the topic, data will be consumed from all data sources documented
      in the run start message
    :param start_time: Get data from now or from start of the last run
    :param stop_time: Stop data_stream at end of run, or not
    """
    """
    Additional info:
    - The `topics` argument is generally unused.
    - Instead, prefer the use of `run_info_topic` where it is possible to go
      and find the last run_start message and use that to get info on what
      topics to listen to.
    - Buffer sizes: it is currently not easy to resize scipp data structures,
      so we choose sensible default buffer sizes. This may need to be tweaked
      in the future.
    - `start_time`: it is possible to go back to the start of the run, even if
      `data_stream()` is started after or during the run. It simply finds the
      start time in the last run_start message. The data can persist on Kafka
      for a significant duration (hours? days?), making this lookup possible.
    """

    validate_buffer_size_args(
        chopper_buffer_size,
        event_buffer_size,
        fast_metadata_buffer_size,
        slow_metadata_buffer_size,
    )

    ctx = mp.get_context("spawn")
    data_queue = ctx.Queue()
    instruction_queue = ctx.Queue()

    # Use "async for" as "yield from" cannot be used in an async function, see
    # https://www.python.org/dev/peps/pep-0525/#asynchronous-yield-from
    async for data_chunk in _data_stream(
        data_queue,
        instruction_queue,
        kafka_broker,
        topics,
        interval,
        event_buffer_size,
        slow_metadata_buffer_size,
        fast_metadata_buffer_size,
        chopper_buffer_size,
        run_info_topic,
        start_time,
        stop_time,
    ):  # noqa: E125
        yield data_chunk


def validate_buffer_size_args(
    chopper_buffer_size,
    event_buffer_size,
    fast_metadata_buffer_size,
    slow_metadata_buffer_size,
):
    for buffer_name, buffer_size in (
        ("event_buffer_size", event_buffer_size),
        ("slow_metadata_buffer_size", slow_metadata_buffer_size),
        ("fast_metadata_buffer_size", fast_metadata_buffer_size),
        ("chopper_buffer_size", chopper_buffer_size),
    ):
        if buffer_size < 1:
            raise ValueError(f"{buffer_name} must be greater than zero")


def _cleanup_queue(queue: Optional[mp.Queue]):
    if queue is not None:
        queue.cancel_join_thread()
        queue.close()
        queue.join_thread()


async def _data_stream(
    data_queue: mp.Queue,
    worker_instruction_queue: mp.Queue,
    kafka_broker: str,
    topics: Optional[List[str]],
    interval: sc.Variable,
    event_buffer_size: int,
    slow_metadata_buffer_size: int,
    fast_metadata_buffer_size: int,
    chopper_buffer_size: int,
    run_info_topic: Optional[str] = None,
    start_at: StartTime = StartTime.NOW,
    end_at: StopTime = StopTime.NEVER,
    query_consumer: Optional["KafkaQueryConsumer"] = None,  # noqa: F821
    consumer_type: ConsumerType = ConsumerType.REAL,
    halt_after_n_data_chunks: int = np.iinfo(np.int32).max,  # noqa: B008
    halt_after_n_warnings: int = np.iinfo(np.int32).max,  # noqa: B008
    test_message_queue: Optional[mp.Queue] = None,  # for tests
    timeout: Optional[sc.Variable] = None,  # for tests
) -> Generator[sc.DataArray, None, None]:
    """
    Main implementation of data stream is extracted to this function so that
    fake consumers can be injected for unit tests
    """

    # Search backwards to find the last run_start message
    try:
        from ._consumer import KafkaQueryConsumer, get_run_start_message
        from ._data_consumption_manager import (
            InstructionType,
            ManagerInstruction,
            data_consumption_manager,
        )
    except ImportError:
        raise ImportError(_missing_dependency_message)

    if topics is None and run_info_topic is None:
        raise ValueError(
            "At least one of 'topics' and 'run_info_topic'" " must be specified"
        )

    # This is defaulted to None in the function signature
    # to avoid it having to be imported earlier
    if query_consumer is None:
        query_consumer = KafkaQueryConsumer(kafka_broker)

    # stream_info contains information on where to look for data and metadata.
    # The data from the start message is yielded as the first chunk of data.
    #
    # TODO: This should, in principle, not look any different from any other
    # chunk of data, right now it seems it may be different?
    # (see https://github.com/scipp/scippneutron/issues/114)
    #
    # Generic data chunk structure: geometry, metadata, and event data are all
    # optional.
    # - first data chunk will most probably contain no event data
    # - subsequent chunks can contain geometry info, if e.g. some pixels have
    #   moved
    # - metadata (e.g. sample environment) might be empty, if values have not
    #   changed
    stream_info = None
    run_id = ""
    run_title = "-"  # for display in widget
    stop_time_ms = None
    n_data_chunks = 0
    if run_info_topic is not None:
        run_start_info = get_run_start_message(run_info_topic, query_consumer)
        run_id = run_start_info.job_id
        run_title = run_start_info.run_name
        # default value for stop_time in message flatbuffer is 0,
        # it means that field has not been populated
        if end_at == StopTime.END_OF_RUN and run_start_info.stop_time != 0:
            stop_time_ms = run_start_info.stop_time
        if topics is None:
            loaded_data, stream_info = load_nexus_json_str(
                run_start_info.nexus_structure, get_start_info=True
            )
            topics = [stream.topic for stream in stream_info]
        else:
            loaded_data, _ = load_nexus_json_str(
                run_start_info.nexus_structure, get_start_info=False
            )
        topics.append(run_info_topic)  # listen for stop run message
        yield loaded_data
        n_data_chunks += 1

    if start_at == StartTime.START_OF_RUN:
        start_time = run_start_info.start_time * sc.Unit("milliseconds")
    else:
        start_time = time.time() * sc.units.s

    # Convert to int and float as easier to pass to mp.Process
    # (sc.Variable would have to be serialised/deserialised)
    start_time_ms = int(sc.to_unit(start_time, "milliseconds").value)
    interval_s = float(sc.to_unit(interval, 's').value)

    # Specify to start the process using the "spawn" method, otherwise
    # on Linux the default is to fork the Python interpreter which
    # is "problematic" in a multithreaded process, this can apparently
    # even cause multiprocessing's own Queue to cause problems when forking.
    # See documentation:
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    # pytorch docs mention Queue problem:
    # https://pytorch.org/docs/stable/notes/multiprocessing.html
    #
    # Note also that daemonising this Process is important so that resources are
    # properly freed when restarting the notebook kernel (or shutting down the
    # notebook entirely).
    data_collect_process = mp.get_context("spawn").Process(
        target=data_consumption_manager,
        args=(
            start_time_ms,
            stop_time_ms,
            run_id,
            topics,
            kafka_broker,
            consumer_type,
            stream_info,
            interval_s,
            event_buffer_size,
            slow_metadata_buffer_size,
            fast_metadata_buffer_size,
            chopper_buffer_size,
            worker_instruction_queue,
            data_queue,
            test_message_queue,
        ),
        daemon=True,
    )
    try:
        data_stream_widget = DataStreamWidget(
            start_time_ms=start_time_ms, stop_time_ms=stop_time_ms, run_title=run_title
        )
        data_collect_process.start()

        # When testing, if something goes wrong, the while loop below can
        # become infinite. So we introduce a timeout.
        if timeout is not None:
            start_timeout = time.time()
            timeout_s = float(sc.to_unit(timeout, 's').value)
        n_warnings = 0
        while (
            data_collect_process.is_alive()
            and n_data_chunks < halt_after_n_data_chunks
            and n_warnings < halt_after_n_warnings
            and not data_stream_widget.stop_requested
        ):
            if timeout is not None and (time.time() - start_timeout) > timeout_s:
                raise TimeoutError("data_stream timed out in test")
            try:
                new_data = data_queue.get_nowait()

                if isinstance(new_data, Warning):
                    # Raise warnings in this process so that they
                    # can be captured in tests
                    warn(new_data)
                    n_warnings += 1
                    continue
                elif isinstance(new_data, StopTimeUpdate):
                    data_stream_widget.set_stop_time(new_data.stop_time_ms)
                    if end_at == StopTime.END_OF_RUN:
                        worker_instruction_queue.put(
                            ManagerInstruction(
                                InstructionType.UPDATE_STOP_TIME, new_data.stop_time_ms
                            )
                        )
                    continue
                n_data_chunks += 1
                yield convert_from_pickleable_dict(new_data)
            except QueueEmpty:
                await asyncio.sleep(0.5 * interval_s)
    finally:
        # Ensure cleanup happens however the loop exits
        worker_instruction_queue.put(ManagerInstruction(InstructionType.STOP_NOW))
        if data_collect_process.is_alive():
            process_halt_timeout_s = 4.0
            data_collect_process.join(process_halt_timeout_s)
        if data_collect_process.is_alive():
            data_collect_process.terminate()
        for queue in (data_queue, worker_instruction_queue, test_message_queue):
            _cleanup_queue(queue)
        data_stream_widget.set_stopped()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import multiprocessing as mp
import threading
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import scipp as sc
from streaming_data_types.eventdata_ev42 import deserialise_ev42
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.logdata_f142 import LogDataInfo, deserialise_f142
from streaming_data_types.run_stop_6s4t import deserialise_6s4t
from streaming_data_types.sample_environment_senv import Location as TimestampLocation
from streaming_data_types.sample_environment_senv import Response as FastSampleEnvData
from streaming_data_types.sample_environment_senv import deserialise_senv
from streaming_data_types.timestamps_tdct import Timestamps, deserialise_tdct

from ..io.nexus._json_nexus import StreamInfo
from ._serialisation import convert_to_pickleable_dict
from ._stop_time import StopTimeUpdate
from ._warnings import BufferSizeWarning, UnknownFlatbufferIdWarning

# The ESS data streaming system uses Google FlatBuffers to serialise
# data to transmit in the Kafka message payload. FlatBuffers uses schemas
# to define the data structure to serialise:
# https://github.com/ess-dmsc/streaming-data-types/
# The ess-streaming-data-types library provides convenience "serialise"
# and "deserialise" functions for each schema:
# https://github.com/ess-dmsc/python-streaming-data-types/
# Each schema is identified by a 4 character string which is included in:
# - the filename of the schema,
# - the serialised buffer as the first 4 bytes,
# - the module name in ess-streaming-data-types.
#
# FlatBuffer schema ids.
# These will change in the future if a breaking change needs to be
# made to the schema.
SLOW_FB_ID = "f142"
FAST_FB_ID = "senv"
CHOPPER_FB_ID = "tdct"
EVENT_FB_ID = "ev42"


def _create_metadata_buffer_array(name: str, unit: str, dtype: Any, buffer_size: int):
    return sc.DataArray(
        sc.zeros(dims=[name], shape=(buffer_size,), unit=unit, dtype=dtype),
        coords={
            "time": sc.zeros(
                dims=[name],
                shape=(buffer_size,),
                unit=sc.Unit("nanoseconds"),
                dtype=np.dtype('datetime64[ns]'),
            )
        },
    )


class _SlowMetadataBuffer:
    """
    Buffer for "slowly" changing metadata from Kafka messages serialised
    according to the flatbuffer schema with id SLOW_FB_ID.
    Typically, the data sources are EPICS PVs with updates published to
    Kafka via the Forwarder (https://github.com/ess-dmsc/forwarder/).
    """

    def __init__(self, stream_info: StreamInfo, buffer_size: int):
        self._buffer_mutex = threading.Lock()
        self._buffer_size = buffer_size
        self._name = stream_info.source_name
        self._buffer_filled_size = 0
        self._data_array = _create_metadata_buffer_array(
            self._name, stream_info.unit, stream_info.dtype, buffer_size
        )

    def append_data(self, log_event: LogDataInfo, emit_data: Callable):
        if self._buffer_filled_size == self._buffer_size:
            emit_data()

        # Each LogDataInfo contains a single value-timestamp pair
        with self._buffer_mutex:
            self._data_array[self._name, self._buffer_filled_size] = log_event.value
            self._data_array.coords["time"][
                self._name, self._buffer_filled_size
            ].value = np.datetime64(log_event.timestamp_unix_ns, 'ns')
            self._buffer_filled_size += 1

    def get_metadata_array(self) -> Tuple[bool, sc.Variable]:
        """
        Copy collected data from the buffer
        """
        with self._buffer_mutex:
            return_array = self._data_array[
                self._name, : self._buffer_filled_size
            ].copy()
            new_data_exists = self._buffer_filled_size != 0
            self._buffer_filled_size = 0
        return new_data_exists, sc.scalar(return_array)


class _FastMetadataBuffer:
    """
    Buffer for "fast" changing metadata from Kafka messages serialised
    according to the flatbuffer schema with id FAST_FB_ID.
    Typical data sources are sample environment apparatus with relatively
    rapidly values which, for efficiency, publish updates directly to Kafka
    rather than via EPICS and the Forwarder.
    """

    def __init__(self, stream_info: StreamInfo, buffer_size: int, data_queue: mp.Queue):
        self._buffer_mutex = threading.Lock()
        self._buffer_size = buffer_size
        self._name = stream_info.source_name
        self._data_queue = data_queue
        self._buffer_filled_size = 0
        self._data_array = _create_metadata_buffer_array(
            self._name, stream_info.unit, stream_info.dtype, buffer_size
        )

    def append_data(self, log_events: FastSampleEnvData, emit_data: Callable):
        # Each FastSampleEnvData contains an array of values and either:
        #  - an array of corresponding timestamps, or
        #  - the timedelta for linearly spaced timestamps
        message_size = log_events.values.size
        if message_size > self._buffer_size:
            self._data_queue.put(
                BufferSizeWarning(
                    "Single message would overflow the fast metadata buffer, "
                    "please restart with a larger buffer size:\n"
                    f"message_size: {message_size}, fast_metadata_buffer_size:"
                    f" {self._buffer_size}. These data have been "
                    f"skipped!"
                )
            )
            return

        if self._buffer_filled_size + message_size > self._buffer_size:
            emit_data()

        def _datetime_to_epoch_ns(input_timestamp: datetime) -> int:
            return int(input_timestamp.timestamp() * 1_000_000_000)

        with self._buffer_mutex:
            self._data_array[
                self._name,
                self._buffer_filled_size : self._buffer_filled_size + message_size,
            ].values = log_events.values
            if log_events.value_ts is not None:
                timestamps = log_events.value_ts
            else:
                # Generate linearly spaced sample timestamps assuming that
                # the message timestamp falls at the start of the sample range
                timestamps = np.arange(0, message_size, dtype=np.int64) * int(
                    log_events.sample_ts_delta
                ) + _datetime_to_epoch_ns(log_events.timestamp)
                if log_events.ts_location == TimestampLocation.Middle:
                    # Shift timestamps so that the message timestamp falls
                    # in the middle of the sample range
                    timestamps = timestamps - int(
                        0.5 * (timestamps[-1] - timestamps[0])
                    )
                elif log_events.ts_location == TimestampLocation.End:
                    # Shift timestamps so that the message timestamp falls at
                    # the end of the sample range
                    timestamps = timestamps - (timestamps[-1] - timestamps[0])
                elif log_events.ts_location == TimestampLocation.Start:
                    pass  # timestamps are already correct
                else:
                    # We have accounted for all enum values so this can only
                    # happen if someone really messed up when serialising the
                    # message or a breaking change was made to the schema
                    # without a new schema id being assigned (which is
                    # against ECDC policy)
                    raise RuntimeError(
                        "Unrecognised timestamp location in fast sample "
                        "environment data message (flatbuffer id: "
                        f"'{FAST_FB_ID}')"
                    )
            self._data_array[
                self._name,
                self._buffer_filled_size : self._buffer_filled_size + message_size,
            ].coords["time"].values = timestamps
            self._buffer_filled_size += message_size

    def get_metadata_array(self) -> Tuple[bool, sc.Variable]:
        """
        Copy collected data from the buffer
        """
        with self._buffer_mutex:
            return_array = self._data_array[
                self._name, : self._buffer_filled_size
            ].copy()
            new_data_exists = self._buffer_filled_size != 0
            self._buffer_filled_size = 0
        return new_data_exists, sc.scalar(return_array)


class _ChopperMetadataBuffer:
    """
    Buffer for chopper top-dead-centre timestamps from Kafka messages
    serialised according to the flatbuffer schema with id CHOPPER_FB_ID.
    """

    def __init__(self, stream_info: StreamInfo, buffer_size: int, data_queue: mp.Queue):
        self._buffer_mutex = threading.Lock()
        self._buffer_size = buffer_size
        self._name = stream_info.source_name
        self._data_queue = data_queue
        self._buffer_filled_size = 0
        self._data_array = sc.zeros(
            dims=[self._name],
            shape=(buffer_size,),
            unit=sc.Unit("nanoseconds"),
            dtype=np.int64,
        )

    def append_data(self, chopper_timestamps: Timestamps, emit_data: Callable):
        # Each Timestamps contains an array of top-dead-centre timestamps
        message_size = chopper_timestamps.timestamps.size
        if message_size > self._buffer_size:
            self._data_queue.put(
                BufferSizeWarning(
                    "Single message would overflow the chopper data buffer, "
                    "please restart with a larger buffer size:\n"
                    f"message_size: {message_size}, chopper_buffer_size:"
                    f" {self._buffer_size}. These data have been "
                    f"skipped!"
                )
            )
            return

        if self._buffer_filled_size + message_size > self._buffer_size:
            emit_data()

        with self._buffer_mutex:
            self._data_array[
                self._name,
                self._buffer_filled_size : self._buffer_filled_size + message_size,
            ].values = chopper_timestamps.timestamps
            self._buffer_filled_size += message_size

    def get_metadata_array(self) -> Tuple[bool, sc.Variable]:
        """
        Copy collected data from the buffer
        """
        with self._buffer_mutex:
            return_array = self._data_array[
                self._name, : self._buffer_filled_size
            ].copy()
            new_data_exists = self._buffer_filled_size != 0
            self._buffer_filled_size = 0
        return new_data_exists, sc.scalar(return_array)


metadata_ids = (SLOW_FB_ID, FAST_FB_ID, CHOPPER_FB_ID)
_MetadataBuffer = Union[
    _FastMetadataBuffer, _SlowMetadataBuffer, _ChopperMetadataBuffer
]


class StreamedDataBuffer:
    """
    This owns the buffers for data consumed from Kafka.
    It periodically emits accumulated data to a queue
    and resets the buffer. If a buffer fills up within the emit time
    interval then data are emitted early.

    TODO: This also owns the metadata buffers. Maybe this should be moved to a
    separate place in the future?
    """

    def __init__(
        self,
        queue: mp.Queue,
        event_buffer_size: int,
        slow_metadata_buffer_size: int,
        fast_metadata_buffer_size: int,
        chopper_buffer_size: int,
        interval_s: float,
        run_id: str,
    ):
        self._buffer_mutex = threading.Lock()
        self._interval_s = interval_s
        self._event_buffer_size = event_buffer_size
        self._slow_metadata_buffer_size = slow_metadata_buffer_size
        self._fast_metadata_buffer_size = fast_metadata_buffer_size
        self._chopper_buffer_size = chopper_buffer_size
        self._current_run_id = run_id
        tof_buffer = sc.zeros(
            dims=['event'],
            shape=[event_buffer_size],
            unit=sc.units.ns,
            dtype=sc.DType.int32,
        )
        id_buffer = sc.zeros(
            dims=['event'],
            shape=[event_buffer_size],
            unit=sc.units.one,
            dtype=sc.DType.int32,
        )
        pulse_times = sc.zeros(
            dims=['event'],
            shape=[event_buffer_size],
            unit=sc.units.ns,
            dtype=sc.DType.int64,
        )
        weights = sc.ones(
            dims=['event'], shape=[event_buffer_size], with_variances=True
        )
        self._events_buffer = sc.DataArray(
            weights,
            coords={
                'tof': tof_buffer,
                'detector_id': id_buffer,
                'pulse_time': pulse_times,
            },
        )
        self._current_event = 0
        self._cancelled = False
        self._notify_cancelled = threading.Condition()
        self._unrecognised_fb_id_count = 0
        self._periodic_emit: Optional[threading.Thread] = None
        self._emit_queue = queue
        # Access metadata buffer by
        # self._metadata_buffers[flatbuffer_id][source_name]
        self._metadata_buffers: Dict[str, Dict[str, _MetadataBuffer]] = {
            flatbuffer_id: {} for flatbuffer_id in metadata_ids
        }

    def init_metadata_buffers(self, stream_info: List[StreamInfo]):
        """
        Create a buffer for each of the metadata sources.
        This is not in the constructor which allows the StreamedDataBuffer
        to be constructed before metadata sources are extracted from the
        run start message, this means the StreamedDataBuffer can be passed
        into data_stream to facilitate unit testing.
        """
        for stream in stream_info:
            if stream.flatbuffer_id == SLOW_FB_ID:
                self._metadata_buffers[stream.flatbuffer_id][
                    stream.source_name
                ] = _SlowMetadataBuffer(stream, self._slow_metadata_buffer_size)
            elif stream.flatbuffer_id == FAST_FB_ID:
                self._metadata_buffers[stream.flatbuffer_id][
                    stream.source_name
                ] = _FastMetadataBuffer(
                    stream, self._fast_metadata_buffer_size, self._emit_queue
                )
            elif stream.flatbuffer_id == CHOPPER_FB_ID:
                self._metadata_buffers[stream.flatbuffer_id][
                    stream.source_name
                ] = _ChopperMetadataBuffer(
                    stream, self._chopper_buffer_size, self._emit_queue
                )
            elif stream.flatbuffer_id == EVENT_FB_ID:
                pass  # detection events, not metadata
            else:
                self._emit_queue.put(
                    UnknownFlatbufferIdWarning(
                        f"Stream in run start message specified flatbuffer id "
                        f"'{stream.flatbuffer_id}' which scippneutron does "
                        f"not know how to deserialise."
                    )
                )

    def start(self):
        self._cancelled = False
        self._unrecognised_fb_id_count = 0
        self._periodic_emit = threading.Thread(target=self._emit_loop)
        self._periodic_emit.start()

    def stop(self):
        self._cancelled = True
        with self._notify_cancelled:
            self._notify_cancelled.notifyAll()
        if self._periodic_emit is not None and self._periodic_emit.is_alive():
            self._periodic_emit.join(5.0)
        self._emit_data()  # flush the buffer

    def _emit_data(self):
        with self._buffer_mutex:
            if self._unrecognised_fb_id_count:
                self._emit_queue.put(
                    UnknownFlatbufferIdWarning(
                        f"Received {self._unrecognised_fb_id_count}"
                        " messages with unrecognised FlatBuffer ids"
                    )
                )
                self._unrecognised_fb_id_count = 0
            new_data = self._events_buffer['event', : self._current_event].copy()
            new_data_exists = self._current_event != 0
            for _, buffers in self._metadata_buffers.items():
                for name, buffer in buffers.items():
                    (new_metadata_exists, metadata_array) = buffer.get_metadata_array()
                    new_data.attrs[name] = metadata_array
                    if new_metadata_exists:
                        new_data_exists = True
            self._current_event = 0
        if new_data_exists:
            self._emit_queue.put(convert_to_pickleable_dict(new_data))

    def _emit_loop(self):
        while not self._cancelled:
            with self._notify_cancelled:
                self._notify_cancelled.wait(timeout=self._interval_s)
            self._emit_data()

    def _handled_event_data(self, new_data: bytes) -> bool:
        try:
            deserialised_data = deserialise_ev42(new_data)
            message_size = deserialised_data.detector_id.size
            if message_size > self._event_buffer_size:
                self._emit_queue.put(
                    BufferSizeWarning(
                        "Single message would overflow the event data buffer, "
                        "please restart with a larger buffer size:\n"
                        f"message_size: {message_size}, event_buffer_size:"
                        f" {self._event_buffer_size}. These data have been "
                        f"skipped!"
                    )
                )
                return True
            # If new data would overfill buffer then emit data
            # currently in buffer first
            if self._current_event + message_size > self._event_buffer_size:
                self._emit_data()
            with self._buffer_mutex:
                frame = self._events_buffer[
                    'event', self._current_event : self._current_event + message_size
                ]
                frame.coords['detector_id'].values = deserialised_data.detector_id
                frame.coords['tof'].values = deserialised_data.time_of_flight
                frame.coords[
                    'pulse_time'
                ].values = deserialised_data.pulse_time * np.ones_like(
                    deserialised_data.time_of_flight
                )
                self._current_event += message_size
        except WrongSchemaException:
            return False
        return True

    def _handled_metadata(
        self, new_data: bytes, source_field_name: str, deserialise: Callable, fb_id: str
    ) -> bool:
        try:
            deserialised_data = deserialise(new_data)
            try:
                self._metadata_buffers[fb_id][
                    getattr(deserialised_data, source_field_name)
                ].append_data(deserialised_data, self._emit_data)
                return True
            except KeyError:
                # Ignore data from unknown source name
                pass
        except WrongSchemaException:
            return False
        return False

    def _handled_stop_run(self, new_data: bytes):
        try:
            stop_run_data = deserialise_6s4t(new_data)
            if stop_run_data.job_id == self._current_run_id:
                self._emit_queue.put(StopTimeUpdate(stop_run_data.stop_time))
            return True
        except WrongSchemaException:
            return False

    def new_data(self, new_data: bytes):
        """
        This is the callback which is given to the consumers.
        """
        if self._handled_event_data(new_data):
            return
        if self._handled_metadata(
            new_data, "source_name", deserialise_f142, SLOW_FB_ID
        ):
            return
        if self._handled_metadata(new_data, "name", deserialise_senv, FAST_FB_ID):
            return
        if self._handled_metadata(new_data, "name", deserialise_tdct, CHOPPER_FB_ID):
            return
        if self._handled_stop_run(new_data):
            return
        # new data were not handled
        self._unrecognised_fb_id_count += 1



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import numpy as np
import pytest
import scipp as sc

import scippneutron as scn


@pytest.fixture
def data_array():
    positions = [[-1, 0, 0], [0, 1, 0], [0, 0, 1]]
    return sc.DataArray(
        sc.Variable(dims=['position', 'tof'], values=np.random.rand(3, 9)),
        coords={
            'position': sc.vectors(dims=['position'], values=positions, unit='m'),
            'source_position': sc.vector(value=np.array([0, 0, -10]), unit='m'),
            'sample_position': sc.vector(value=np.array([0, 1, 0]), unit='m'),
        },
    )


def test_position(data_array):
    assert sc.identical(scn.position(data_array), data_array.coords['position'])


def test_source_position(data_array):
    assert sc.identical(
        scn.source_position(data_array),
        sc.vector(value=np.array([0, 0, -10]), unit='m'),
    )


def test_sample_position(data_array):
    assert sc.identical(
        scn.sample_position(data_array), sc.vector(value=np.array([0, 1, 0]), unit='m')
    )


def test_L1(data_array):
    assert sc.identical(scn.L1(data_array), sc.scalar(np.sqrt(1 + 100), unit='m'))


def test_L2(data_array):
    assert sc.identical(
        scn.L2(data_array),
        sc.array(dims=['position'], values=np.sqrt([1 + 1, 0, 1 + 1]), unit='m'),
    )


def test_Ltotal_scatter(data_array):
    assert sc.identical(
        scn.Ltotal(data_array, scatter=True), scn.L1(data_array) + scn.L2(data_array)
    )


def test_Ltotal_no_scatter(data_array):
    assert sc.identical(
        scn.Ltotal(data_array, scatter=False),
        sc.array(dims=['position'], values=np.sqrt([1 + 100, 1 + 100, 121]), unit='m'),
    )


def test_two_theta(data_array):
    # Bare-bones test because computation of 2theta is complicated,
    # and it is tested separately elsewhere.
    two_theta = scn.two_theta(data_array)
    assert two_theta.unit == 'rad'
    assert two_theta.sizes == scn.position(data_array).sizes



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import datetime
import multiprocessing as mp
import platform
import sys
import warnings
from cmath import isclose
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy as np
import pytest
import scipp as sc

from scippneutron._utils import get_attrs
from scippneutron.data_streaming._consumer_type import ConsumerType
from scippneutron.data_streaming._warnings import (
    BufferSizeWarning,
    UnknownFlatbufferIdWarning,
)

sys.path.insert(0, str(Path(__file__).resolve().parent / 'io'))
from nexus_helpers import EventData, Log, NexusBuilder, Stream  # noqa: E402

if platform.system() == "Darwin":
    pytest.skip(
        "Streaming tests disabled on MacOS because they tend to hang"
        " indefinitely for unknown reasons",
        allow_module_level=True,
    )

try:
    import streaming_data_types  # noqa: F401
    from confluent_kafka import KafkaError, TopicPartition  # noqa: F401
    from streaming_data_types.eventdata_ev42 import serialise_ev42  # noqa: E402
    from streaming_data_types.logdata_f142 import serialise_f142
    from streaming_data_types.run_start_pl72 import serialise_pl72
    from streaming_data_types.sample_environment_senv import Location, serialise_senv
    from streaming_data_types.timestamps_tdct import serialise_tdct

    from scippneutron.data_streaming._consumer import RunStartError
    from scippneutron.data_streaming.data_stream import _data_stream  # noqa: E402
    from scippneutron.data_streaming.data_stream import StopTime
except ImportError:
    pytest.skip("Kafka or Serialisation module is unavailable", allow_module_level=True)


class FakeKafkaError:
    def __init__(self, kafka_error_code: int):
        self._error_code = kafka_error_code

    def code(self):
        return self._error_code

    def __str__(self):
        return f"FakeKafkaError: code {self._error_code}"


class FakeMessage:
    def __init__(
        self, payload: bytes, error_code: Optional[int] = None, timestamp: int = 0
    ):
        self._payload = payload
        self._error_code = error_code
        self._timestamp = timestamp

    def value(self):
        return self._payload

    def error(self) -> Optional[KafkaError]:
        if self._error_code is not None:
            return FakeKafkaError(self._error_code)
        return None

    def timestamp(self) -> Tuple[None, int]:
        return None, self._timestamp


class FakeQueryConsumer:
    def __init__(
        self,
        instrument_name: str = "",
        low_and_high_offset: Tuple[int, int] = (2, 10),
        streams: List[Stream] = None,
        start_time: Optional[Union[int, datetime.datetime]] = None,
        stop_time: Optional[Union[int, datetime.datetime]] = None,
        nexus_structure: Optional[str] = None,
    ):
        self._instrument_name = instrument_name
        self._low_and_high_offset = low_and_high_offset
        self._streams = streams
        self.queried_topics = []
        self.queried_timestamp = None
        self._start_time = start_time
        self._stop_time = stop_time
        if self._start_time is None:
            self._start_time = datetime.datetime.now()
        if nexus_structure is None:
            builder = NexusBuilder()
            builder.add_instrument(self._instrument_name)
            if self._streams is not None:
                for stream in self._streams:
                    builder.add_stream(stream)
            self._nexus_structure = builder.json_string
        else:
            self._nexus_structure = nexus_structure

    @staticmethod
    def assign(partitions: List[TopicPartition]):
        pass

    def get_watermark_offsets(self, partition: TopicPartition) -> Tuple[int, int]:
        return self._low_and_high_offset

    def get_topic_partitions(
        self, topic: str, offset: int = -1
    ) -> List[TopicPartition]:
        self.queried_topics.append(topic)
        return [TopicPartition(topic, partition=0, offset=offset)]

    def poll(self, timeout=2.0) -> FakeMessage:
        return FakeMessage(
            serialise_pl72(
                "",
                "",
                start_time=self._start_time,
                stop_time=self._stop_time,
                nexus_structure=self._nexus_structure,
            )
        )

    def seek(self, partition: TopicPartition):
        pass

    def offsets_for_times(self, partitions: List[TopicPartition]):
        self.queried_timestamp = partitions[0].offset
        return partitions


# Short time to use for buffer emit and data_stream interval in tests
# pass or fail fast!
SHORT_TEST_INTERVAL = 100.0 * sc.Unit('milliseconds')
# Small buffer of 20 events is sufficient for the tests
TEST_BUFFER_SIZE = 20

TEST_STREAM_ARGS = {
    "kafka_broker": "broker",
    "topics": ["topic"],
    "interval": SHORT_TEST_INTERVAL,
    "event_buffer_size": TEST_BUFFER_SIZE,
    "slow_metadata_buffer_size": TEST_BUFFER_SIZE,
    "fast_metadata_buffer_size": TEST_BUFFER_SIZE,
    "chopper_buffer_size": TEST_BUFFER_SIZE,
    "consumer_type": ConsumerType.FAKE,
    "timeout": 10.0 * sc.units.s,
}
# "timeout" arg: if something gets broken then this makes sure the
# test will not get stuck in the _data_stream loop indefinitely.
# A TimeoutError is raised by _data_stream if the timeout occurs.


@pytest.fixture(scope="function")
def queues():
    # Specify to start the process using the "spawn" method, otherwise
    # on Linux the default is to fork the Python interpreter which
    # is "problematic" in a multithreaded process, in our case the use of
    # asyncio means the process is multithreaded.
    # See documentation:
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    ctx = mp.get_context("spawn")
    return ctx.Queue(), ctx.Queue(), ctx.Queue()


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_single_event_message(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    time_of_flight = np.array([1.0, 2.0, 3.0])
    detector_ids = np.array([4, 5, 6])
    test_message = FakeMessage(
        serialise_ev42("detector", 0, 0, time_of_flight, detector_ids)
    )
    test_message_queue.put(test_message)

    reached_assert = False
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        halt_after_n_data_chunks=1,
        test_message_queue=test_message_queue,
        query_consumer=FakeQueryConsumer(),
        **TEST_STREAM_ARGS,
    ):
        assert np.allclose(data.coords['tof'].values, time_of_flight)
        reached_assert = True
    assert reached_assert


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_event_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    first_tof = np.array([1.0, 2.0, 3.0])
    first_detector_ids = np.array([4, 5, 6])
    first_test_message = serialise_ev42("detector", 0, 0, first_tof, first_detector_ids)
    second_tof = np.array([1.0, 2.0, 3.0])
    second_detector_ids = np.array([4, 5, 6])
    second_test_message = serialise_ev42(
        "detector", 0, 0, second_tof, second_detector_ids
    )
    test_message_queue.put(FakeMessage(first_test_message))
    test_message_queue.put(FakeMessage(second_test_message))

    reached_asserts = False
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        halt_after_n_data_chunks=1,
        test_message_queue=test_message_queue,
        query_consumer=FakeQueryConsumer(),
        **TEST_STREAM_ARGS,
    ):
        expected_tofs = np.concatenate((first_tof, second_tof))
        assert np.allclose(data.coords['tof'].values, expected_tofs)
        expected_ids = np.concatenate((first_detector_ids, second_detector_ids))
        assert np.array_equal(data.coords['detector_id'].values, expected_ids)
        reached_asserts = True
    assert reached_asserts


@pytest.mark.asyncio
async def test_warn_if_unrecognised_message_was_encountered(queues):
    warnings.filterwarnings("error")
    data_queue, worker_instruction_queue, test_message_queue = queues
    # First 4 bytes of the message payload are the FlatBuffer schema identifier
    # "abcd" does not correspond to a FlatBuffer schema for data
    # that scipp is interested in
    test_message = b"abcd0000"

    with pytest.warns(UnknownFlatbufferIdWarning):
        test_message_queue.put(FakeMessage(test_message))
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            halt_after_n_warnings=1,
            test_message_queue=test_message_queue,
            query_consumer=FakeQueryConsumer(),
            **TEST_STREAM_ARGS,
        ):
            test_message_queue.put(FakeMessage(test_message))


@pytest.mark.asyncio
async def test_warn_on_buffer_size_exceeded_by_single_message(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size_2_events = 2
    time_of_flight = np.array([1.0, 2.0, 3.0])
    detector_ids = np.array([4, 5, 6])
    test_message = serialise_ev42("detector", 0, 0, time_of_flight, detector_ids)

    test_steam_args = TEST_STREAM_ARGS.copy()
    test_steam_args["event_buffer_size"] = buffer_size_2_events

    with pytest.warns(BufferSizeWarning):
        test_message_queue.put(FakeMessage(test_message))
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            halt_after_n_warnings=1,
            test_message_queue=test_message_queue,
            query_consumer=FakeQueryConsumer(),
            **test_steam_args,
        ):
            test_message_queue.put(FakeMessage(test_message))


@pytest.mark.asyncio
async def test_data_returned_when_buffer_size_exceeded_by_event_messages(queues):
    # Messages cumulatively exceed the buffer size, data_stream
    # will return multiple chunks of data to clear the buffer
    # between messages.
    data_queue, worker_instruction_queue, test_message_queue = queues
    first_tof = np.array([1.0, 2.0, 3.0])
    first_detector_ids = np.array([4, 5, 6])
    first_test_message = serialise_ev42("detector", 0, 0, first_tof, first_detector_ids)
    second_tof = np.array([7.0, 8.0, 9.0])
    second_detector_ids = np.array([4, 5, 6])
    second_test_message = serialise_ev42(
        "detector", 0, 0, second_tof, second_detector_ids
    )

    # Event data buffer size is 5, so the second message
    # will not fit in the buffer
    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["event_buffer_size"] = 5
    test_stream_args["run_info_topic"] = "run_info_topic"

    reached_asserts = False
    n_chunks = 0
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        halt_after_n_data_chunks=3,
        test_message_queue=test_message_queue,
        query_consumer=FakeQueryConsumer(),
        **test_stream_args,
    ):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_test_message))
            test_message_queue.put(FakeMessage(second_test_message))
        elif n_chunks == 1:
            # Contain event data from first message
            assert np.allclose(data.coords['tof'].values, first_tof)
        elif n_chunks == 2:
            # Contain event data from second message
            assert np.allclose(data.coords['tof'].values, second_tof)
            reached_asserts = True

        n_chunks += 1
    assert reached_asserts


@pytest.mark.asyncio
async def test_data_are_loaded_from_run_start_message(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    reached_assert = False
    test_instrument_name = "DATA_STREAM_TEST"
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        halt_after_n_data_chunks=0,
        test_message_queue=test_message_queue,
        query_consumer=FakeQueryConsumer(test_instrument_name),
        **TEST_STREAM_ARGS,
    ):
        assert data["instrument_name"].value == test_instrument_name
        reached_assert = True
    assert reached_assert


@pytest.mark.asyncio
async def test_error_raised_if_no_run_start_message_available(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # Low and high offset are the same value, indicates there are
    # no messages available in the partition
    low_and_high_offset = (0, 0)
    with pytest.raises(RunStartError):
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=run_info_topic,
            halt_after_n_data_chunks=0,
            test_message_queue=test_message_queue,
            query_consumer=FakeQueryConsumer(test_instrument_name, low_and_high_offset),
            **TEST_STREAM_ARGS,
        ):
            pass


@pytest.mark.asyncio
async def test_error_if_both_topics_and_run_start_topic_not_specified(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # At least one of "topics" and "run_start_topic" must be specified
    with pytest.raises(ValueError):
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=None,
            halt_after_n_data_chunks=0,
            **test_stream_args,
            query_consumer=FakeQueryConsumer(),
            test_message_queue=test_message_queue,
        ):
            pass


@pytest.mark.asyncio
async def test_specified_topics_override_run_start_message_topics(queues):
    # If "topics" argument is specified then they should be used, even if
    # a run start topic is provided
    data_queue, worker_instruction_queue, test_message_queue = queues
    test_topics = ["whiting", "snail", "porpoise"]
    topic_in_run_start_message = "test_topic"
    test_streams = [Stream("/entry", topic_in_run_start_message)]
    query_consumer = FakeQueryConsumer(streams=test_streams)
    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = test_topics
    async for _ in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=None,
        query_consumer=query_consumer,
        halt_after_n_data_chunks=0,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        pass
    assert not query_consumer.queried_topics, (
        "Expected specified topics" " to be used and none queried"
    )


@pytest.mark.asyncio
async def test_data_stream_returns_metadata(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    # There are currently 3 schemas for metadata, they have flatbuffer ids
    # f142, senv and tdct
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(
            f"/entry/{f142_log_name}",
            "f142_topic",
            f142_source_name,
            "f142",
            "double",
            "m",
        ),
        Stream(
            f"/entry/{senv_log_name}",
            "senv_topic",
            senv_source_name,
            "senv",
            "double",
            "m",
        ),
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct"),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        halt_after_n_data_chunks=2,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            f142_value = 26.1236
            f142_timestamp = 123456  # ns after epoch
            f142_test_message = serialise_f142(
                f142_value, f142_source_name, f142_timestamp
            )
            test_message_queue.put(FakeMessage(f142_test_message))
            senv_values = np.array([26, 127, 52])
            senv_timestamp_ns = 123000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(
                senv_timestamp_ns * 1e-9, datetime.timezone.utc
            )
            senv_time_between_samples = 100  # ns
            senv_test_message = serialise_senv(
                senv_source_name,
                -1,
                senv_timestamp,
                senv_time_between_samples,
                0,
                senv_values,
                Location.Start,
            )
            test_message_queue.put(FakeMessage(senv_test_message))
            tdct_timestamps = np.array([1234, 2345, 3456])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps)
            test_message_queue.put(FakeMessage(tdct_test_message))

        n_chunks += 1

    assert isclose(
        get_attrs(data_from_stream)[f142_source_name].value.values[0], f142_value
    )
    assert get_attrs(data_from_stream)[f142_source_name].value.coords['time'].values[
        0
    ] == np.array(f142_timestamp, dtype=np.dtype('datetime64[ns]'))
    assert np.array_equal(
        get_attrs(data_from_stream)[senv_source_name].value.values, senv_values
    )
    senv_expected_timestamps = np.array(
        [
            senv_timestamp_ns,
            senv_timestamp_ns + senv_time_between_samples,
            senv_timestamp_ns + (2 * senv_time_between_samples),
        ],
        dtype=np.dtype('datetime64[ns]'),
    )
    assert np.array_equal(
        get_attrs(data_from_stream)[senv_source_name].value.coords['time'].values,
        senv_expected_timestamps,
    )
    assert np.array_equal(
        get_attrs(data_from_stream)[tdct_source_name].value.values,
        tdct_timestamps,
    )


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_slow_metadata_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    streams = [
        Stream(
            f"/entry/{f142_log_name}",
            "f142_topic",
            f142_source_name,
            "f142",
            "double",
            "m",
        ),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        halt_after_n_data_chunks=2,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            f142_value_1 = 26.1236
            f142_timestamp_1 = 123456  # ns after epoch
            f142_test_message = serialise_f142(
                f142_value_1, f142_source_name, f142_timestamp_1
            )
            test_message_queue.put(FakeMessage(f142_test_message))
            f142_value_2 = 2.725
            f142_timestamp_2 = 234567  # ns after epoch
            f142_test_message = serialise_f142(
                f142_value_2, f142_source_name, f142_timestamp_2
            )
            test_message_queue.put(FakeMessage(f142_test_message))

        n_chunks += 1

    assert np.allclose(
        get_attrs(data_from_stream)[f142_source_name].value.values,
        np.array([f142_value_1, f142_value_2]),
    )
    assert np.array_equal(
        get_attrs(data_from_stream)[f142_source_name].value.coords['time'].values,
        np.array(
            [f142_timestamp_1, f142_timestamp_2], dtype=np.dtype('datetime64[ns]')
        ),
    )


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_fast_metadata_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    streams = [
        Stream(
            f"/entry/{senv_log_name}",
            "senv_topic",
            senv_source_name,
            "senv",
            "double",
            "m",
        ),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        halt_after_n_data_chunks=2,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            senv_values_1 = np.array([26, 127, 52])
            senv_timestamp_ns_1 = 123000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(
                senv_timestamp_ns_1 * 1e-9, datetime.timezone.utc
            )
            senv_time_between_samples = 100  # ns
            senv_test_message = serialise_senv(
                senv_source_name,
                -1,
                senv_timestamp,
                senv_time_between_samples,
                0,
                senv_values_1,
                Location.Start,
            )
            test_message_queue.put(FakeMessage(senv_test_message))
            senv_values_2 = np.array([3832, 324, 3])
            senv_timestamp_ns_2 = 234000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(
                senv_timestamp_ns_2 * 1e-9, datetime.timezone.utc
            )
            senv_test_message = serialise_senv(
                senv_source_name,
                -1,
                senv_timestamp,
                senv_time_between_samples,
                0,
                senv_values_2,
                Location.Start,
            )
            test_message_queue.put(FakeMessage(senv_test_message))

        n_chunks += 1

    assert np.array_equal(
        get_attrs(data_from_stream)[senv_source_name].value.values,
        np.concatenate((senv_values_1, senv_values_2)),
    )
    senv_expected_timestamps_1 = np.array(
        [
            senv_timestamp_ns_1,
            senv_timestamp_ns_1 + senv_time_between_samples,
            senv_timestamp_ns_1 + (2 * senv_time_between_samples),
        ],
        dtype=np.dtype('datetime64[ns]'),
    )
    senv_expected_timestamps_2 = np.array(
        [
            senv_timestamp_ns_2,
            senv_timestamp_ns_2 + senv_time_between_samples,
            senv_timestamp_ns_2 + (2 * senv_time_between_samples),
        ],
        dtype=np.dtype('datetime64[ns]'),
    )
    assert np.array_equal(
        get_attrs(data_from_stream)[senv_source_name].value.coords['time'].values,
        np.concatenate((senv_expected_timestamps_1, senv_expected_timestamps_2)),
    )


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_chopper_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        halt_after_n_data_chunks=2,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            tdct_timestamps_1 = np.array([1234, 2345, 3456])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps_1)
            test_message_queue.put(FakeMessage(tdct_test_message))
            tdct_timestamps_2 = np.array([4567, 5678, 6789])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps_2)
            test_message_queue.put(FakeMessage(tdct_test_message))

        n_chunks += 1

    assert np.array_equal(
        get_attrs(data_from_stream)[tdct_source_name].value.values,
        np.concatenate((tdct_timestamps_1, tdct_timestamps_2)),
    )


@pytest.mark.asyncio
async def test_data_stream_warns_if_fast_metadata_message_exceeds_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 2
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    streams = [
        Stream(
            f"/entry/{senv_log_name}",
            "senv_topic",
            senv_source_name,
            "senv",
            "double",
            "m",
        ),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["fast_metadata_buffer_size"] = buffer_size
    with pytest.warns(BufferSizeWarning):
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=run_info_topic,
            query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
            halt_after_n_warnings=1,
            **test_stream_args,
            test_message_queue=test_message_queue,
        ):
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.

            # 3 values but buffer size is only 2!
            senv_values = np.array([26, 127, 52])
            senv_timestamp_ns = 123000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(
                senv_timestamp_ns * 1e-9, datetime.timezone.utc
            )
            senv_time_between_samples = 100  # ns
            senv_test_message = serialise_senv(
                senv_source_name,
                -1,
                senv_timestamp,
                senv_time_between_samples,
                0,
                senv_values,
                Location.Start,
            )

            test_message_queue.put(FakeMessage(senv_test_message))


@pytest.mark.asyncio
async def test_data_stream_warns_if_single_chopper_message_exceeds_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 2
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["chopper_buffer_size"] = buffer_size
    with pytest.warns(BufferSizeWarning):
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=run_info_topic,
            query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
            halt_after_n_warnings=1,
            **test_stream_args,
            test_message_queue=test_message_queue,
        ):
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.

            # 3 values but buffer size is only 2!
            tdct_timestamps = np.array([1234, 2345, 3456])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps)

            test_message_queue.put(FakeMessage(tdct_test_message))


@pytest.mark.asyncio
async def test_data_returned_if_multiple_slow_metadata_msgs_exceed_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    streams = [
        Stream(
            f"/entry/{f142_log_name}",
            "f142_topic",
            f142_source_name,
            "f142",
            "double",
            "m",
        ),
    ]

    first_f142_value = 26.1236
    f142_timestamp = 123456  # ns after epoch
    first_message = serialise_f142(first_f142_value, f142_source_name, f142_timestamp)
    second_f142_value = 62.721
    second_message = serialise_f142(second_f142_value, f142_source_name, f142_timestamp)

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["slow_metadata_buffer_size"] = 1
    test_stream_args["topics"] = None
    n_chunks = 0
    reached_asserts = False
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        halt_after_n_data_chunks=3,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_message))
            test_message_queue.put(FakeMessage(second_message))
        elif n_chunks == 1:
            # Contains data from first message
            assert isclose(
                get_attrs(data)[f142_source_name].value.values[0],
                first_f142_value,
            )
        elif n_chunks == 2:
            # Contains data from second message
            assert isclose(
                get_attrs(data)[f142_source_name].value.values[0],
                second_f142_value,
            )
            reached_asserts = True
        n_chunks += 1

    assert reached_asserts


@pytest.mark.asyncio
async def test_data_returned_if_multiple_fast_metadata_msgs_exceed_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 4
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    streams = [
        Stream(
            f"/entry/{senv_log_name}",
            "senv_topic",
            senv_source_name,
            "senv",
            "double",
            "m",
        ),
    ]

    first_senv_values = np.array([26, 127, 52])
    second_senv_values = np.array([72, 94, 1])
    senv_timestamp_ns = 123000  # ns after epoch
    senv_timestamp = datetime.datetime.fromtimestamp(
        senv_timestamp_ns * 1e-9, datetime.timezone.utc
    )
    senv_time_between_samples = 100  # ns
    first_message = serialise_senv(
        senv_source_name,
        -1,
        senv_timestamp,
        senv_time_between_samples,
        0,
        first_senv_values,
        Location.Start,
    )
    second_message = serialise_senv(
        senv_source_name,
        -1,
        senv_timestamp,
        senv_time_between_samples,
        0,
        second_senv_values,
        Location.Start,
    )

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["fast_metadata_buffer_size"] = buffer_size
    n_chunks = 0
    reached_asserts = False
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        halt_after_n_data_chunks=3,
        **test_stream_args,
        test_message_queue=test_message_queue,
    ):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_message))
            test_message_queue.put(FakeMessage(second_message))
        elif n_chunks == 1:
            # Contains data from first message
            assert np.array_equal(
                get_attrs(data)[senv_source_name].value.values, first_senv_values
            )
        elif n_chunks == 2:
            # Contains data from second message
            assert np.array_equal(
                get_attrs(data)[senv_source_name].value.values, second_senv_values
            )
            reached_asserts = True
        n_chunks += 1
    assert reached_asserts


@pytest.mark.asyncio
async def test_data_returned_if_multiple_chopper_msgs_exceed_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 4
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    tdct_timestamps_1 = np.array([1234, 2345, 3456])  # ns
    first_tdct_message = serialise_tdct(tdct_source_name, tdct_timestamps_1)
    tdct_timestamps_2 = np.array([4567, 5678, 6789])  # ns
    second_tdct_message = serialise_tdct(tdct_source_name, tdct_timestamps_2)

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["chopper_buffer_size"] = buffer_size

    n_chunks = 0
    reached_asserts = False
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(test_instrument_name, streams=streams),
        **test_stream_args,
        halt_after_n_data_chunks=3,
        test_message_queue=test_message_queue,
    ):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_tdct_message))
            test_message_queue.put(FakeMessage(second_tdct_message))
        elif n_chunks == 1:
            # Contains data from first message
            assert np.array_equal(
                get_attrs(data)[tdct_source_name].value.values, tdct_timestamps_1
            )
        elif n_chunks == 2:
            # Contains data from second message
            assert np.array_equal(
                get_attrs(data)[tdct_source_name].value.values, tdct_timestamps_2
            )
            reached_asserts = True
        n_chunks += 1
    assert reached_asserts


@pytest.mark.asyncio
async def test_passes_with_missing_datasets_if_group_contains_stream(queues):
    # Create NeXus description for run start message which contains
    # an NXlog which contains no datasets but does have a Stream
    # source for the data
    builder = NexusBuilder()
    test_instrument_name = "DATA_STREAM_TEST"
    builder.add_instrument(test_instrument_name)
    builder.add_log(Log("log", None))
    builder.add_event_data(EventData(None, None, None, None))
    builder.add_stream(Stream("/entry/log"))
    builder.add_stream(Stream("/entry/events_0"))
    nexus_structure = builder.json_string

    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    reached_assert = False

    with warnings.catch_warnings(record=True):
        warnings.filterwarnings("ignore", message='Failed to load')
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=run_info_topic,
            query_consumer=FakeQueryConsumer(
                test_instrument_name, nexus_structure=nexus_structure
            ),
            **TEST_STREAM_ARGS,
            halt_after_n_data_chunks=0,
            test_message_queue=test_message_queue,
        ):
            reached_assert = True
            break
        assert reached_assert


@pytest.mark.asyncio
async def test_data_stream_times_out(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # "timeout" is only for use in tests, ensures that if something
    # gets broken then tests will not get stuck in the _data_stream
    # loop indefinitely.
    test_stream_args["timeout"] = 2.0 * sc.units.s
    timed_out = False
    try:
        async for _ in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=run_info_topic,
            query_consumer=FakeQueryConsumer(test_instrument_name),
            **test_stream_args,
            test_message_queue=test_message_queue,
            end_at=StopTime.END_OF_RUN,
        ):
            # Do nothing until it times out
            pass
    except TimeoutError:
        timed_out = True
    assert timed_out


@pytest.mark.asyncio
async def test_stream_loop_exits_if_stop_time_and_end_of_partition_reached(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # System time is already after this stop time so the stream will stop
    # as soon as it sees the end of partition or a message with a
    # timestamp after the stop time
    stop_time_in_past = datetime.datetime(2017, 11, 28, 23, 55, 59, 342380)
    n_chunks = 0
    async for _ in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(
            test_instrument_name, stop_time=stop_time_in_past
        ),
        **test_stream_args,
        test_message_queue=test_message_queue,
        end_at=StopTime.END_OF_RUN,
    ):
        if n_chunks == 0:
            # Tell consumer it has reached the end of the partition
            # Consumer will stop, data_consumption_manager will see the
            # consumer has stopped and stop the buffer and the
            # _data_stream will exit.
            # A TimeoutError would occur if the functionality is broken.
            test_message_queue.put(FakeMessage(b"", KafkaError._PARTITION_EOF))
        n_chunks += 1


@pytest.mark.asyncio
async def test_stream_loop_exits_if_stop_time_reached_and_later_message_seen(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    streams = [
        Stream(
            f"/entry/{f142_log_name}",
            "f142_topic",
            f142_source_name,
            "f142",
            "double",
            "m",
        ),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # System time is already after this stop time so the stream will stop
    # as soon as it sees the end of partition or a message with a
    # timestamp after the stop time
    stop_time_in_past = datetime.datetime(2017, 11, 28, 23, 55, 59, 342380)
    n_chunks = 0
    async for data in _data_stream(
        data_queue,
        worker_instruction_queue,
        run_info_topic=run_info_topic,
        query_consumer=FakeQueryConsumer(
            test_instrument_name, stop_time=stop_time_in_past, streams=streams
        ),
        **test_stream_args,
        test_message_queue=test_message_queue,
        end_at=StopTime.END_OF_RUN,
    ):
        if n_chunks == 0:
            # Publish a message with a timestamp before the stop time
            f142_value_1 = 26.1236
            timestamp_before_stop_dt = datetime.datetime(2017, 11, 28, 23, 55, 50, 0)
            # Convert to integer nanoseconds
            # (for timestamp in message payload)
            timestamp_before_stop_ns = int(
                timestamp_before_stop_dt.timestamp() * 1_000_000_000
            )
            # Convert to integer milliseconds
            # (for Kafka message header)
            timestamp_before_stop_ms = int(timestamp_before_stop_dt.timestamp() * 1_000)
            f142_test_message = serialise_f142(
                f142_value_1, f142_source_name, timestamp_before_stop_ns
            )
            test_message_queue.put(
                FakeMessage(f142_test_message, timestamp=timestamp_before_stop_ms)
            )
        elif n_chunks == 1:
            # The data from the first message will be returned
            assert np.allclose(
                get_attrs(data)[f142_source_name].value.values,
                np.array([f142_value_1]),
            )
            assert np.array_equal(
                get_attrs(data)[f142_source_name].value.coords['time'].values,
                np.array([timestamp_before_stop_ns], dtype=np.dtype('datetime64[ns]')),
            )

            # Publish message with timestamp after stop time, this will trigger
            # the consumer to stop and data_stream to exit.
            # A TimeoutError would occur if the functionality is broken.
            f142_value_2 = 2.725
            timestamp_after_stop_dt = datetime.datetime(2017, 11, 28, 23, 56, 50, 0)
            timestamp_after_stop_ns = int(
                timestamp_after_stop_dt.timestamp() * 1_000_000_000
            )
            timestamp_after_stop_ms = int(timestamp_after_stop_dt.timestamp() * 1_000)
            f142_test_message = serialise_f142(
                f142_value_2, f142_source_name, timestamp_after_stop_ns
            )
            test_message_queue.put(
                FakeMessage(f142_test_message, timestamp=timestamp_after_stop_ms)
            )

        n_chunks += 1



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""Graphs for computing beamline parameters from positions.

See :mod:`scippneutron.conversion.beamline` for definitions
of the quantities used here.
"""

from typing import Callable, Dict

from .. import beamline as _kernels

Graph = Dict[str, Callable]


def incident_beam() -> Graph:
    """Graph for computing 'incident_beam'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return {'incident_beam': _kernels.straight_incident_beam}


def scattered_beam() -> Graph:
    """Graph for computing 'scattered_beam'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return {'scattered_beam': _kernels.straight_scattered_beam}


def two_theta() -> Graph:
    """Graph for computing the scattering angle 'two_theta'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    graph = beamline(scatter=True)
    for node in ['L1', 'L2', 'Ltotal']:
        del graph[node]
    return graph


def L1() -> Graph:
    """Graph for computing the primary path length 'L1'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return {'incident_beam': _kernels.straight_incident_beam, 'L1': _kernels.L1}


def L2() -> Graph:
    """Graph for computing the secondary path length 'L2'.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    return {'scattered_beam': _kernels.straight_scattered_beam, 'L2': _kernels.L2}


def Ltotal(scatter: bool) -> Graph:
    """Graph for computing 'Ltotal'.

    Parameters
    ----------
    scatter:
        If True, a graph for scattering from ``sample_position`` is
        returned, else a graph without scattering.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    graph = beamline(scatter=scatter)
    if not scatter:
        return graph
    del graph['two_theta']
    return graph


_NO_SCATTER_GRAPH_BEAMLINE = {
    'Ltotal': _kernels.total_straight_beam_length_no_scatter,
}

_SCATTER_GRAPH_BEAMLINE = {
    'incident_beam': _kernels.straight_incident_beam,
    'scattered_beam': _kernels.straight_scattered_beam,
    'L1': _kernels.L1,
    'L2': _kernels.L2,
    'two_theta': _kernels.two_theta,
    'Ltotal': _kernels.total_beam_length,
}


def beamline(scatter: bool) -> Graph:
    """Graph defining a straight beamline geometry.

    This can be used as part of transformation graphs that require, e.g., scattering
    angles (``two_theta``) or flight path lengths (``L1`` and ``L2``).

    Parameters
    ----------
    scatter:
        If True, a graph for scattering from ``sample_position`` is
        returned, else a graph without scattering.

    Returns
    -------
    :
        A dict defining a coordinate transformation graph.
    """
    if scatter:
        return dict(_SCATTER_GRAPH_BEAMLINE)
    else:
        return dict(_NO_SCATTER_GRAPH_BEAMLINE)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# flake8: noqa
"""Neutron scattering toolkit built using scipp for Data Reduction.

ScippNeutron is a generic (as in 'usable by different facilities')
package for data processing in neutron scattering.
It provides coordinate transformations, file I/O, and technique-specific tools.

See the online documentation for user guides and the API reference:
https://scipp.github.io/scippneutron/
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .beamline_components import (
    position,
    source_position,
    sample_position,
    incident_beam,
    scattered_beam,
    Ltotal,
    L1,
    L2,
    two_theta,
)
from .core import convert
from .mantid import (
    from_mantid,
    array_from_mantid,
    to_mantid,
    load_with_mantid,
    load,
    fit,
)
from .instrument_view import instrument_view
from .io.nexus.load_nexus import load_nexus, load_nexus_json
from .data_streaming.data_stream import data_stream
from . import atoms
from . import data

del importlib



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from datetime import datetime
from typing import Optional


def _unix_ms_to_datetime(unix_ms: int) -> datetime:
    return datetime.fromtimestamp(unix_ms / 1000.0)


class DataStreamWidget:
    def __init__(
        self,
        start_time_ms: Optional[int] = None,
        stop_time_ms: Optional[int] = None,
        run_title: Optional[str] = None,
    ):
        import ipywidgets as widgets
        from IPython.display import display

        self._stop_button = widgets.ToggleButton(description="Stop stream")
        self._stop_button.observe(self._on_button_clicked, "value")
        self._title = widgets.Label("-")
        self._start_time = widgets.Label("-")
        self._stop_time = widgets.Label("-")
        self._time_format = "%m/%d/%Y %H:%M:%S"

        if run_title is not None:
            self.set_title(run_title)
        if start_time_ms is not None:
            self.set_start_time(start_time_ms)
        if stop_time_ms is not None and stop_time_ms != 0:
            # 0 is the default in the run start message flatbuffer
            # if stop time field is not populated
            self.set_stop_time(stop_time_ms)

        display(
            widgets.HBox(
                [
                    self._stop_button,
                    widgets.HBox([widgets.Label('Run title:'), self._title]),
                    widgets.HBox([widgets.Label('Start time:'), self._start_time]),
                    widgets.HBox([widgets.Label('Stop time:'), self._stop_time]),
                ]
            )
        )

    @staticmethod
    def _on_button_clicked(b):
        b["owner"].description = "Stopping..."
        b["owner"].disabled = True

    def set_start_time(self, start_time_ms: int):
        self._start_time.value = _unix_ms_to_datetime(start_time_ms).strftime(
            self._time_format
        )

    def set_stop_time(self, stop_time_ms: int):
        self._stop_time.value = _unix_ms_to_datetime(stop_time_ms).strftime(
            self._time_format
        )

    def set_title(self, new_run_title: str):
        self._title.value = new_run_title

    @property
    def stop_requested(self) -> bool:
        return self._stop_button.value

    def set_stopped(self):
        self._stop_button.description = "Stopped"
        self._stop_button.disabled = True



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Matthew Jones
import json
from contextlib import contextmanager
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Dict, Optional, Set, Tuple, Union
from warnings import warn

import h5py
import numpy as np
import scipp as sc
import scippnexus as snx
from scipp.binning import make_binned
from scipp.core.util import VisibleDeprecationWarning
from scippnexus.v1 import NXroot
from scippnexus.v1.nxobject import NexusStructureError, NXobject
from scippnexus.v1.nxtransformations import TransformationError

from ..._utils import get_attrs
from ._json_nexus import JSONGroup, StreamInfo, contains_stream, get_streams_info
from ._nexus import ScippData


class BadSource(Exception):
    """
    Raise if something is wrong with data source which
    prevents it being used. Warn the user.
    """

    pass


class SkipSource(Exception):
    """
    Raise to abort using the data source, do not
    warn the user.
    """

    pass


def add_position_and_transforms_to_data(
    data: Union[sc.DataArray, sc.Dataset],
    transform_name: str,
    position_name: str,
    base_position_name: str,
    transforms: sc.Variable,
    positions: sc.Variable,
):
    if isinstance(data, sc.DataArray):
        coords = data.coords
        attrs = get_attrs(data)
    else:
        coords = data
        attrs = data

    if transforms is None:
        coords[position_name] = positions
        attrs[base_position_name] = positions
    elif isinstance(transforms, sc.Variable):
        # If transform is not time-dependent.
        coords[position_name] = transforms * positions
        attrs[base_position_name] = positions
        attrs[transform_name] = sc.scalar(value=transforms)
    else:
        coords[base_position_name] = positions
        coords[transform_name] = sc.scalar(value=transforms)


@contextmanager
def _open_if_path(file_in: Union[str, Path, h5py.File]):
    """
    Open if file path is provided,
    otherwise yield the existing h5py.File object
    """
    if isinstance(file_in, (str, Path)):
        with h5py.File(file_in, "r", libver='latest', swmr=True) as nexus_file:
            yield nexus_file
    else:
        yield file_in


def _load_instrument_name(instruments: Dict[str, NXobject]) -> Dict:
    instrument = next(iter(instruments.values()))
    if len(instruments) > 1:
        warn(
            f"More than one NXinstrument found in file, "
            f"loading name from {instrument.name} only"
        )
    if (name := instrument.get("name")) is not None:
        return {"instrument_name": sc.scalar(name[()])}
    return {}


def _load_title(entry: NXobject) -> Dict:
    if (title := entry.get('title')) is not None:
        return {"experiment_title": sc.scalar(title[()])}
    return {}


def _load_start_and_end_time(entry: NXobject) -> Dict:
    times = {}
    for time in ["start_time", "end_time"]:
        if (dataset := entry.get(time)) is not None:
            times[time] = sc.scalar(dataset[()])
    return times


def load_nexus(
    data_file: Union[str, Path, h5py.File], root: str = "/", quiet=True
) -> Optional[ScippData]:
    """
    Load a NeXus file and return required information.

    :param data_file: path of NeXus file containing data to load
    :param root: path of group in file, only load data from the subtree of
      this group
    :param quiet: if False prints some details of what is being loaded

    Usage example:
      data = sc.neutron.load_nexus('PG3_4844_event.nxs')
    """
    warn(
        "`load_nexus` is deprecated and will be removed in version 24.03, "
        "please switch to using ScippNexus.",
        VisibleDeprecationWarning,
    )
    start_time = timer()

    with _open_if_path(data_file) as nexus_file:
        loaded_data = _load_data(nexus_file, root, quiet)

    if not quiet:
        print("Total time:", timer() - start_time)
    return loaded_data


def _origin(unit) -> sc.Variable:
    return sc.vector(value=[0, 0, 0], unit=unit)


def _depends_on_to_position(obj) -> Union[None, sc.Variable]:
    if (transform := obj.get('depends_on')) is not None:
        if (
            isinstance(transform, (str, sc.DataArray))
            or transform.dtype == sc.DType.DataArray
        ):
            return None  # cannot compute position if bad transform or time-dependent
        else:
            if transform.dtype == sc.DType.rotation3:
                return transform * _origin('m')
            else:
                return transform.to(unit='m') * _origin('m')


def _monitor_to_canonical(monitor):
    if isinstance(monitor, sc.DataGroup):
        return monitor
    if monitor.bins is not None:
        monitor.bins.coords['tof'] = monitor.bins.coords.pop('event_time_offset')
        monitor.bins.coords['detector_id'] = monitor.bins.coords.pop('event_id')
        monitor.bins.coords['pulse_time'] = sc.bins_like(
            monitor, fill_value=monitor.coords.pop('event_time_zero')
        )
        da = sc.DataArray(
            sc.broadcast(monitor.data.bins.concat('pulse'), dims=['tof'], shape=[1])
        )
    else:
        da = monitor.copy(deep=False)
    if (position := _depends_on_to_position(monitor.coords)) is not None:
        da.coords['position'] = position
    return da


def _zip_pixel_offset(da: sc.DataArray) -> sc.DataArray:
    """Remove [xyz]_pixel_offset and replace them by single 'pixel_offset' coord.

    Returns unchanged data array if x_pixel_offset is not found.
    """
    if 'x_pixel_offset' not in da.coords:
        return da
    x = da.coords.pop('x_pixel_offset')
    offset = sc.zeros(dims=da.dims, shape=da.shape, unit=x.unit, dtype=sc.DType.vector3)
    offset.fields.x = x.to(dtype='float64', copy=False)
    if (y := da.coords.pop('y_pixel_offset', None)) is not None:
        offset.fields.y = y.to(dtype='float64', unit=x.unit, copy=False)
    if (z := da.coords.pop('z_pixel_offset', None)) is not None:
        offset.fields.z = z.to(dtype='float64', unit=x.unit, copy=False)
    da.coords['pixel_offset'] = offset.rename_dims(dict(zip(offset.dims, da.dims)))
    return da


def _by_nx_class(group) -> Dict[str, Dict[str, 'NXobject']]:
    from scippnexus.v1.nxobject import _nx_class_registry

    classes = {name: [] for name in _nx_class_registry()}

    def _match_nx_class(_, node):
        if not hasattr(node, 'shape'):
            if (nx_class := node.attrs.get('NX_class')) is not None:
                if not isinstance(nx_class, str):
                    nx_class = nx_class.decode('UTF-8')
                if nx_class in _nx_class_registry():
                    classes[nx_class].append(node)

    group._group.visititems(_match_nx_class)

    out = {}
    for nx_class, groups in classes.items():
        names = [group.name.split('/')[-1] for group in groups]
        if len(names) != len(set(names)):  # fall back to full path if duplicate
            names = [group.name for group in groups]
        out[nx_class] = {n: group._make(g) for n, g in zip(names, groups)}
    return out


def _load_data(
    nexus_file: Union[h5py.File, Dict], root: Optional[str], quiet: bool
) -> Optional[ScippData]:
    """
    Main implementation for loading data is extracted to this function so that
    in-memory data can be used for unit tests.
    """
    root = NXroot(nexus_file if root is None else nexus_file[root])
    classes = _by_nx_class(root)

    if len(classes['NXentry']) > 1:
        # We can't sensibly load from multiple NXentry, for example each
        # could could contain a description of the same detector bank
        # and lead to problems with clashing detector ids etc
        raise RuntimeError(
            f"More than one NXentry group in file, use 'root' argument "
            "to specify which to load data from, for example"
            f"{__name__}('my_file.nxs', '/entry_2')"
        )

    # In the following, we map the file structure onto a partially flattened in-memory
    # structure. This behavior is quite error prone and cumbersome and will probably
    # disappear in this form. We therefore keep this length code directly in this
    # function to provide an overview and facility future refactoring steps.
    detectors = classes.get('NXdetector', {})
    loaded_detectors = []
    for group in detectors.values():
        try:
            det = group[()]
            if isinstance(det, sc.DataGroup):
                raise NexusStructureError(f"Failed to load NXdetector {group.name}")
            det = _zip_pixel_offset(det)
            det = det.flatten(to='detector_id')
            det.bins.coords['tof'] = det.bins.coords.pop('event_time_offset')
            det.bins.coords['pulse_time'] = det.bins.coords.pop('event_time_zero')
            if 'detector_number' in det.coords:
                det.coords['detector_id'] = det.coords.pop('detector_number')
            elif 'pixel_id' in det.coords:
                det.coords['detector_id'] = det.coords.pop('pixel_id')
            elif 'spectrum_index' in det.coords:
                det.coords['detector_id'] = det.coords.pop('spectrum_index')
            else:
                raise KeyError(
                    "Found neither of detector_number, pixel_id, or spectrum_index."
                )
            if 'pixel_offset' in det.coords:
                add_position_and_transforms_to_data(
                    data=det,
                    transform_name="position_transformations",
                    position_name="position",
                    base_position_name="base_position",
                    positions=det.coords.pop('pixel_offset'),
                    transforms=det.coords.pop('depends_on', None),
                )
            loaded_detectors.append(det)
        except (
            BadSource,
            SkipSource,
            NexusStructureError,
            KeyError,
            sc.DTypeError,
            ValueError,
            IndexError,
        ) as e:
            if not contains_stream(group._group):
                warn(f"Skipped loading {group.name} due to:\n{e}")

    no_event_data = True
    loaded_data = {}

    # If no event data are found, make a Dataset and add the metadata as
    # Dataset entries. Otherwise, make a DataArray.
    if len(loaded_detectors):
        no_event_data = False
        loaded_data = sc.concat(loaded_detectors, 'detector_id')
    elif len(detectors) == 0:
        # If there are no NXdetector groups, load NXevent_data directly
        loaded_events = []
        for group in classes.get('NXevent_data', {}).values():
            try:
                events = group[()]
                if isinstance(events, sc.DataGroup):
                    raise NexusStructureError(f"Failed to load NXdetector {group.name}")
                events.coords['pulse_time'] = events.coords.pop('event_time_zero')
                events.bins.coords['tof'] = events.bins.coords.pop('event_time_offset')
                events.bins.coords['detector_id'] = events.bins.coords.pop('event_id')
                det_min = events.bins.coords['detector_id'].min().value
                det_max = events.bins.coords['detector_id'].max().value
                if len(events.bins.constituents['data']) != 0:
                    # See scipp/scipp#2490
                    det_id = sc.arange('detector_id', det_min, det_max + 1, unit=None)
                    events = make_binned(
                        events, groups=[det_id], erase=['pulse', 'bank']
                    )
                loaded_events.append(events)
            except (BadSource, SkipSource, NexusStructureError, IndexError) as e:
                if not contains_stream(group._group):
                    warn(f"Skipped loading {group.name} due to:\n{e}")
        if len(loaded_events):
            no_event_data = False
            loaded_data = sc.concat(loaded_events, 'detector_id')

    if not no_event_data:
        # Add single tof bin
        loaded_data = sc.DataArray(
            loaded_data.data.fold(
                dim='detector_id', sizes={'detector_id': -1, 'tof': 1}
            ),
            coords=dict(loaded_data.coords.items()),
            attrs=dict(get_attrs(loaded_data).items()),
        )
        tof_min = loaded_data.bins.coords['tof'].min().to(dtype='float64')
        tof_max = loaded_data.bins.coords['tof'].max().to(dtype='float64')
        tof_max.value = np.nextafter(tof_max.value, float("inf"))
        loaded_data.coords['tof'] = sc.concat([tof_min, tof_max], 'tof')

    def add_metadata(metadata: Dict[str, sc.Variable]):
        for key, value in metadata.items():
            if isinstance(loaded_data, sc.DataArray):
                get_attrs(loaded_data)[key] = value
            else:
                loaded_data[key] = value

    if entries := classes['NXentry']:
        entry = next(iter(entries.values()))
        add_metadata(_load_title(entry))
        add_metadata(_load_start_and_end_time(entry))
    if instruments := classes['NXinstrument']:
        add_metadata(_load_instrument_name(instruments))

    def load_and_add_metadata(groups, process=lambda x: x):
        items = {}
        loaded_groups = []
        for name, group in groups.items():
            try:
                items[name] = sc.scalar(process(group[()]))
                loaded_groups.append(name)
            except (
                BadSource,
                SkipSource,
                TransformationError,
                sc.DimensionError,
                sc.UnitError,
                NexusStructureError,
                KeyError,
                ValueError,
            ) as e:
                if not contains_stream(group._group):
                    warn(f"Skipped loading {group.name} due to:\n{e}")
        add_metadata(items)
        return loaded_groups

    load_and_add_metadata(classes.get('NXdisk_chopper', {}))
    load_and_add_metadata(classes.get('NXlog', {}))
    load_and_add_metadata(classes.get('NXmonitor', {}), _monitor_to_canonical)
    for name, tag in {'sample': 'NXsample', 'source': 'NXsource'}.items():
        comps = classes.get(tag, {})
        comps = load_and_add_metadata(comps)
        attrs = loaded_data if isinstance(loaded_data, dict) else get_attrs(loaded_data)
        coords = loaded_data if isinstance(loaded_data, dict) else loaded_data.coords
        for comp_name in comps:
            comp = attrs[comp_name].value
            if (position := _depends_on_to_position(comp)) is not None:
                coords[f'{comp_name}_position'] = position
            elif (distance := comp.get('distance')) is not None:
                if not isinstance(distance, sc.Variable):
                    distance = sc.scalar(distance, unit=None)
                coords[f'{comp_name}_position'] = sc.vector(
                    value=[0, 0, distance.value], unit=distance.unit
                )
            elif name == 'sample':
                coords[f'{comp_name}_position'] = _origin('m')

    # Return None if we have an empty dataset at this point
    if no_event_data and not loaded_data.keys():
        return None
    elif isinstance(loaded_data, sc.DataArray):
        return loaded_data
    return sc.Dataset(loaded_data)


def load_nexus_json_str(
    json_template: str,
    get_start_info: bool = False,
) -> Tuple[Optional[ScippData], Optional[sc.Variable], Optional[Set[StreamInfo]]]:
    """
    Use this function for testing so that file io is not required
    """
    # We do not use cls to convert value lists to sc.Variable at this
    # point because we do not know what dimension names to use here
    loaded_json = json.loads(json_template)
    streams = None
    if get_start_info:
        streams = get_streams_info(loaded_json)
    group = JSONGroup(loaded_json)
    return _load_data(group, None, True), streams


def load_nexus_json(json_filename: str) -> Optional[ScippData]:
    with open(json_filename, 'r') as json_file:
        json_string = json_file.read()
    loaded_data, _ = load_nexus_json_str(json_string)
    return loaded_data


def json_nexus_group(
    json_dict: dict[str, Any], *, definitions: Optional[Dict[str, type]] = None
) -> snx.Group:
    """Parse a JSON dictionary into a NeXus group.

    Parameters
    ----------
    json_dict:
        ``dict`` containing a NeXus structure as JSON.
    definitions:
        ScippNexus application definitions.
        When not given, the default definitions are used.

    Returns
    -------
    :
        A NeXus group that can be used for loading data as if it were
        loaded from a file with :class:`scippnexus.File`.
    """
    return snx.Group(
        JSONGroup(json_dict),
        definitions=definitions if definitions is not None else snx.base_definitions(),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import pytest
import scipp as sc

import scippneutron as scn

try:
    import mantid.kernel as kernel
    import mantid.simpleapi as sapi
except ImportError:
    pytestmark = pytest.mark.skip('Mantid framework is unavailable')
    sapi = None
    kernel = None


@pytest.fixture
def in_ws():
    ws = sapi.Load(Filename=scn.data.get_path("CNCS_51936_event.nxs"), StoreInADS=False)
    return ws


@pytest.fixture
def in_da(in_ws):
    return scn.mantid.from_mantid(in_ws)['data'].astype('float64')


def test_histogram_events(in_ws, in_da):
    from scipp.utils.comparison import isnear

    out_ws = sapi.Rebin(
        InputWorkspace=in_ws,
        Params=[0, 10, 1000],
        PreserveEvents=False,
        StoreInADS=False,
    )
    out_mantid = scn.mantid.from_mantid(out_ws)['data']
    out_scipp = in_da.hist(
        tof=sc.linspace('tof', 0, 1000, num=101, dtype='float64', unit='us')
    )
    assert isnear(out_scipp, out_mantid, rtol=1e-15 * sc.units.one)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import numpy as np
import scipp as sc

from scippneutron.conversion import beamline


def test_straight_incident_beam():
    source_position = sc.vectors(
        dims=['siti'], values=[[40, 80, 20], [30, 10, 50]], unit='mm'
    )
    sample_position = sc.vector([7, 5, 3], unit='mm')
    incident_beam = beamline.straight_incident_beam(
        source_position=source_position, sample_position=sample_position
    )
    assert sc.allclose(
        incident_beam,
        sc.vectors(dims=['siti'], values=[[-33, -75, -17], [-23, -5, -47]], unit='mm'),
    )


def test_straight_scattered_beam():
    position = sc.vectors(dims=['on'], values=[[6, 3, 9], [8, 3, 1]], unit='km')
    sample_position = sc.vector([0.4, 0.6, 0.2], unit='km')
    scattered_beam = beamline.straight_scattered_beam(
        position=position, sample_position=sample_position
    )
    assert sc.allclose(
        scattered_beam,
        sc.vectors(dims=['on'], values=[[5.6, 2.4, 8.8], [7.6, 2.4, 0.8]], unit='km'),
    )


def test_L1():
    incident_beam = sc.vectors(
        dims=['inc'], values=[[0.5, 1.0, 1.5], [-0.3, 0.6, -0.9]], unit='um'
    )
    L1 = beamline.L1(incident_beam=incident_beam)
    assert sc.allclose(
        L1, sc.array(dims=['inc'], values=[1.870828693386, 1.122497216032], unit='um')
    )


def test_L2():
    scattered_beam = sc.vectors(
        dims=['scat'], values=[[11, 22, 33], [95, 84, 73]], unit='am'
    )
    L2 = beamline.L2(scattered_beam=scattered_beam)
    assert sc.allclose(
        L2,
        sc.array(dims=['scat'], values=[41.158231254513, 146.321563687653], unit='am'),
    )


def test_total_beam_length():
    L1 = sc.scalar(5.134, unit='cm')
    L2 = sc.array(dims=['secondary'], values=[3.14159, 42.0, 999.0], unit='cm')
    Ltotal = beamline.total_beam_length(L1=L1, L2=L2)
    assert sc.allclose(
        Ltotal,
        sc.array(dims=['secondary'], values=[8.27559, 47.134, 1004.134], unit='cm'),
    )


def test_total_straight_beam_length_no_scatter():
    position = sc.vectors(dims=['po'], values=[[1, 2, 3], [4, 5, 6]], unit='m')
    source_position = sc.vector([0.1, 0.2, 0.3], unit='m')
    Ltotal = beamline.total_straight_beam_length_no_scatter(
        source_position=source_position, position=position
    )
    assert sc.allclose(
        Ltotal, sc.array(dims=['po'], values=[3.367491648096, 8.410707461325], unit='m')
    )


def test_two_theta_arbitrary_values():
    incident_beam = sc.vector([0.564, 1.2, -10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[13, 24, 35], [51, -42, 33]], unit='m'
    )
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    assert sc.allclose(
        two_theta,
        sc.array(dims=['beam'], values=[2.352629742382, 2.061447408052], unit='rad'),
    )


def test_two_theta_parallel_beams():
    incident_beam = sc.vector([1.0, 0.2, 0.0], unit='mm')
    scattered_beam = sc.array(dims=['beam'], values=[1.0, 2.0, 0.1]) * incident_beam
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    assert sc.allclose(two_theta, sc.zeros(dims=['beam'], shape=[3], unit='rad'))


def test_two_theta_anti_parallel_beams():
    incident_beam = sc.vector([10.2, -0.8, 4.1], unit='mm')
    scattered_beam = sc.array(dims=['beam'], values=[-2.1, -31.0, -1.0]) * incident_beam
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    assert sc.allclose(
        two_theta, sc.full(value=np.pi, dims=['beam'], shape=[3], unit='rad')
    )


def test_two_theta_orthogonal_beams():
    incident_beam = sc.vector([0.1, 0.0, 10.0], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[0.0, -2.1, 0.0], [10.0, 123.0, -0.1]], unit='m'
    )
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    assert sc.allclose(
        two_theta, sc.full(value=np.pi / 2, dims=['beam'], shape=[2], unit='rad')
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Convert a Scipp DataArray to a picklable dictionary and back.
Can be used to move DataArrays between multiprocessing.Process.
"""

from typing import Any, Dict

import numpy as np
import scipp as sc

_scipp_containers = ("DataArray", "DataSet", "Variable")


def convert_to_pickleable_dict(data: sc.DataArray) -> Dict:
    data_dict = sc.to_dict(data)

    def _unit_and_dtype_to_str(d: Any):
        for k, v in d.items():
            if isinstance(v, dict):
                _unit_and_dtype_to_str(v)
            elif k == "unit":
                d[k] = str(v)
            elif k == "dtype":
                d[k] = str(v)
                if any(
                    scipp_container_type == d[k]
                    for scipp_container_type in _scipp_containers
                ):
                    d["values"] = sc.to_dict(d["values"])
                    _unit_and_dtype_to_str(d["values"])

    _unit_and_dtype_to_str(data_dict)
    return data_dict


def convert_from_pickleable_dict(data_dict: Dict) -> sc.DataArray:
    def convert_from_str_unit_and_dtype(d):
        delete_dtype = False
        for k, v in d.items():
            if isinstance(v, dict):
                convert_from_str_unit_and_dtype(v)
                if k not in ("attrs", "masks", "coords"):
                    if {"coords", "data"}.issubset(set(v.keys())):
                        # from_dict does not work with nested DataArrays,
                        # so we have to manually construct DataArrays here.
                        d[k] = sc.DataArray(
                            coords=v["coords"], data=v["data"], attrs=v["attrs"]
                        )
                    else:
                        try:
                            if any(
                                scipp_container_type in str(v["dtype"])
                                for scipp_container_type in _scipp_containers
                            ):
                                del v["dtype"]
                        except KeyError:
                            pass
                        d[k] = sc.from_dict(v)
            else:
                if k == "dtype":
                    try:
                        d[k] = np.dtype(v)
                    except TypeError:
                        if v == "string":
                            d[k] = sc.DType.string
                        elif any(
                            scipp_container_type in k
                            for scipp_container_type in _scipp_containers
                        ):
                            delete_dtype = True
        # Delete now, not while looping through dictionary
        if delete_dtype:
            del d["dtype"]

    convert_from_str_unit_and_dtype(data_dict)
    return sc.DataArray(
        data=data_dict["data"], coords=data_dict["coords"], attrs=data_dict["attrs"]
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import multiprocessing as mp
from dataclasses import dataclass
from enum import Enum
from queue import Empty as QueueEmpty
from typing import List, Optional

from ..io.nexus.load_nexus import StreamInfo
from ._consumer import (
    all_consumers_stopped,
    create_consumers,
    start_consumers,
    stop_consumers,
)
from ._consumer_type import ConsumerType
from ._data_buffer import StreamedDataBuffer


class InstructionType(Enum):
    STOP_NOW = 1
    UPDATE_STOP_TIME = 2


@dataclass(frozen=True)
class ManagerInstruction:
    type: InstructionType
    stop_time_ms: Optional[int] = None  # milliseconds from unix epoch


def data_consumption_manager(
    start_time_ms: int,
    stop_time_ms: Optional[int],
    run_id: str,
    topics: List[str],
    kafka_broker: str,
    consumer_type: ConsumerType,
    stream_info: Optional[List[StreamInfo]],
    interval_s: float,
    event_buffer_size: int,
    slow_metadata_buffer_size: int,
    fast_metadata_buffer_size: int,
    chopper_buffer_size: int,
    worker_instruction_queue: mp.Queue,
    data_queue: mp.Queue,
    test_message_queue: Optional[mp.Queue],
):
    """
    Starts and stops buffers and data consumers which collect data and
    send them back to the main process via a queue.

    All input args must be mp.Queue or pickleable as this function is launched
    as a multiprocessing.Process.
    """
    buffer = StreamedDataBuffer(
        data_queue,
        event_buffer_size,
        slow_metadata_buffer_size,
        fast_metadata_buffer_size,
        chopper_buffer_size,
        interval_s,
        run_id,
    )

    if stream_info is not None:
        buffer.init_metadata_buffers(stream_info)

    consumers = create_consumers(
        start_time_ms,
        stop_time_ms,
        set(topics),
        kafka_broker,
        consumer_type,
        buffer.new_data,
        test_message_queue,
    )

    start_consumers(consumers)
    buffer.start()

    while not all_consumers_stopped(consumers):
        try:
            instruction = worker_instruction_queue.get(timeout=0.5)
            if instruction.type == InstructionType.STOP_NOW:
                stop_consumers(consumers)
            elif instruction.type == InstructionType.UPDATE_STOP_TIME:
                for consumer in consumers:
                    consumer.update_stop_time(instruction.stop_time_ms)
        except QueueEmpty:
            pass
        except (ValueError, OSError):
            # Queue has been closed, stop worker
            stop_consumers(consumers)

    buffer.stop()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
class UnknownFlatbufferIdWarning(Warning):
    pass


class BufferSizeWarning(Warning):
    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

from typing import Callable, Dict, Tuple, Union

import scipp as sc

from ..conversion import graph as _graphs


def _inelastic_scatter_graph(energy_mode):
    inelastic_graph_factory = {
        'direct_inelastic': _graphs.tof.direct_inelastic,
        'indirect_inelastic': _graphs.tof.indirect_inelastic,
    }
    return {
        **_graphs.beamline.beamline(scatter=True),
        **inelastic_graph_factory[energy_mode](start='tof'),
    }


def _reachable_by(target, graph):
    return any(
        target == targets if isinstance(targets, str) else target in targets
        for targets in graph.keys()
    )


def _elastic_scatter_graph(origin, target):
    scatter_graph_kinematics = _graphs.beamline.beamline(scatter=True)
    if _reachable_by(target, scatter_graph_kinematics):
        return dict(scatter_graph_kinematics)
    return {**scatter_graph_kinematics, **_graphs.tof.elastic(origin)}


def _scatter_graph(origin, target, energy_mode):
    graph = (
        _elastic_scatter_graph(origin, target)
        if energy_mode == 'elastic'
        else _inelastic_scatter_graph(energy_mode)
    )
    return graph


def conversion_graph(
    origin: str, target: str, scatter: bool, energy_mode: str
) -> Dict[Union[str, Tuple[str]], Callable]:
    """
    Get a conversion graph for given parameters.

    The graph can be used with `scipp.transform_coords`.

    :param origin: Name of the input coordinate.
    :param target: Name of the output coordinate.
    :param scatter: Choose whether to use scattering or non-scattering conversions.
    :param energy_mode: Select if energy is conserved. One of `'elastic'`,
                        `'direct_inelastic'`, `'indirect_inelastic'`.
    :return: Conversion graph.
    :seealso: :py:func:`scippneutron.convert`,
              :py:func:`scippneutron.deduce_conversion_graph`.
    """

    # Results are copied to ensure users do not modify the global dictionaries.
    if scatter:
        return dict(_scatter_graph(origin, target, energy_mode))
    else:
        return {
            **_graphs.beamline.beamline(scatter=False),
            **_graphs.tof.kinematic(start='tof'),
        }


def _find_inelastic_inputs(data):
    return [name for name in ('incident_energy', 'final_energy') if name in data.coords]


def _deduce_energy_mode(data, origin, target):
    inelastic_inputs = _find_inelastic_inputs(data)
    if target == 'energy_transfer':
        if len(inelastic_inputs) > 1:
            raise RuntimeError(
                "Data contains coords for incident *and* final energy, cannot have "
                "both for inelastic scattering."
            )
        if len(inelastic_inputs) == 0:
            raise RuntimeError(
                "Data contains neither coords for incident nor for final energy, this "
                "does not appear to be inelastic-scattering data, cannot convert to "
                "energy transfer."
            )
        return {
            'incident_energy': 'direct_inelastic',
            'final_energy': 'indirect_inelastic',
        }[inelastic_inputs[0]]

    if 'energy' in (origin, target):
        if inelastic_inputs:
            raise RuntimeError(
                f"Data contains coords for inelastic scattering "
                f"({inelastic_inputs}) but conversion with elastic energy requested. "
                f"This is not implemented."
            )
    return 'elastic'


def deduce_conversion_graph(
    data: Union[sc.DataArray, sc.Dataset], origin: str, target: str, scatter: bool
) -> Dict[Union[str, Tuple[str]], Callable]:
    """
    Get the conversion graph used by :py:func:`scippneutron.convert`
    when called with identical arguments.

    :param data: Input data.
    :param origin: Name of the input coordinate.
    :param target: Name of the output coordinate.
    :param scatter: Choose whether to use scattering or non-scattering conversions.
    :return: Conversion graph.
    :seealso: :py:func:`scippneutron.convert`, :py:func:`scippneutron.conversion_graph`.
    """
    return conversion_graph(
        origin, target, scatter, _deduce_energy_mode(data, origin, target)
    )


def convert(
    data: Union[sc.DataArray, sc.Dataset], origin: str, target: str, scatter: bool
) -> Union[sc.DataArray, sc.Dataset]:
    """
    Perform a unit conversion from the given origin unit to target.
    See the documentation page on "Coordinate Transformations"
    (https://scipp.github.io/scippneutron/user-guide/coordinate-transformations.html)
    for more details.

    :param data: Input data.
    :param origin: Name of the input coordinate.
    :param target: Name of the output coordinate.
    :param scatter: Choose whether to use scattering or non-scattering conversions.
    :return: A new scipp.DataArray or scipp.Dataset with the new coordinate.
    :seealso: :py:func:`scippneutron.deduce_conversion_graph` and
              :py:func:`scippneutron.conversion_graph` to inspect
              the possible conversions.
    """

    graph = deduce_conversion_graph(data, origin, target, scatter)

    try:
        converted = data.transform_coords(target, graph=graph)
    except KeyError as err:
        if err.args[0] == target:
            raise RuntimeError(
                f"No viable conversion from '{origin}' to '{target}' "
                f"with scatter={scatter}."
            )
        raise RuntimeError(
            f"Missing coordinate '{err.args[0]}' for conversion "
            f"from '{origin}' to '{target}'"
        ) from None

    return converted
